@@ -42,18 +42,32 @@ defmodule Application do
       end
 
   In Mix projects, the environment of the application and its dependencies can
-  be overridden via the `config/config.exs` file. For example, someone using
-  your application can override its `:db_host` environment variable as follows:
+  be overridden via the `config/config.exs` and `config/runtime.exs` files. The
+  former is loaded at build-time, before your code compiles, and the latter at
+  runtime, just before your app starts. For example, someone using your application
+  can override its `:db_host` environment variable as follows:
 
       import Config
       config :my_app, :db_host, "db.local"
 
+  See the "Configuration" section in the `Mix` module for more information.
   You can also change the application environment dynamically by using functions
-  such as `put_env/3` and `delete_env/2`. However, as a rule of thumb, each application
-  is responsible for its own environment. Please do not use the functions in this
-  module for directly accessing or modifying the environment of other applications.
+  such as `put_env/3` and `delete_env/2`.
 
-  ### Compile-time environment
+  > Note: The config files `config/config.exs` and `config/runtime.exs`
+  > are rarely used by libraries. Libraries typically define their environment
+  > in the `def application` function of their `mix.exs`. Configuration files
+  > are rather used by applications to configure their libraries.
+
+  > Note: Each application is responsible for its own environment. Do not
+  > use the functions in this module for directly accessing or modifying
+  > the environment of other applications. Whenever you change the application
+  > environment, Elixir's build tool will only recompile the files that
+  > belong to that application. So if you read the application environment
+  > of another application, there is a chance you will be depending on
+  > outdated configuration, as your file won't be recompiled as it changes.
+
+  ## Compile-time environment
 
   In the previous example, we read the application environment at runtime:
 
@@ -71,36 +85,48 @@ defmodule Application do
   will only be read when `MyApp.DBClient` effectively starts. While reading
   the application environment at runtime is the preferred approach, in some
   rare occasions you may want to use the application environment to configure
-  the compilation of a certain project. This is often done by calling `get_env/3`
-  outside of a function:
+  the compilation of a certain project. However, if you try to access
+  `Application.fetch_env!/2` outside of a function:
 
       defmodule MyApp.DBClient do
-        @db_host Application.get_env(:my_app, :db_host, "db.local")
+        @db_host Application.fetch_env!(:my_app, :db_host)
 
         def start_link() do
           SomeLib.DBClient.start_link(host: @db_host)
         end
       end
 
-  This approach has one big limitation: if you change the value of the
-  application environment after the code is compiled, the value used at
-  runtime is not going to change! For example, if your `config/runtime.exs`
-  has:
+  You might see warnings and errors:
 
-      config :my_app, :db_host, "db.production"
+      warning: Application.fetch_env!/2 is discouraged in the module body,
+      use Application.compile_env/3 instead
+        iex:3: MyApp.DBClient
 
-  This value will have no effect as the code was compiled to connect to "db.local",
-  which is mostly likely unavailable in the production environment.
+      ** (ArgumentError) could not fetch application environment :db_host
+      for application :my_app because the application was not loaded nor
+      configured
 
-  For those reasons, reading the application environment at runtime should be the
-  first choice. However, if you really have to read the application environment
-  during compilation, we recommend you to use `compile_env/3` instead:
+  This happens because, when defining modules, the application environment
+  is not yet available. Luckily, the warning tells us how to solve this
+  issue, by using `Application.compile_env/3` instead:
 
-      @db_host Application.compile_env(:my_app, :db_host, "db.local")
+      defmodule MyApp.DBClient do
+        @db_host Application.compile_env(:my_app, :db_host, "db.local")
 
-  By using `compile_env/3`, tools like Mix will store the values used during
-  compilation and compare the compilation values with the runtime values whenever
-  your system starts, raising an error in case they differ.
+        def start_link() do
+          SomeLib.DBClient.start_link(host: @db_host)
+        end
+      end
+
+  The difference here is that `compile_env` expects the default value to be
+  given as an argument, instead of using the `def application` function of
+  your `mix.exs`. Furthermore, by using `compile_env/3`, tools like Mix will
+  store the values used during compilation and compare the compilation values
+  with the runtime values whenever your system starts, raising an error in
+  case they differ.
+
+  In any case, compile-time environments should be avoided. Whenever possible,
+  reading the application environment at runtime should be the first choice.
 
   ## The application callback module
 
@@ -163,7 +189,7 @@ defmodule Application do
   In the sections above, we have configured an application in the
   `application/0` section of the `mix.exs` file. Ultimately, Mix will use
   this configuration to create an [*application resource
-  file*](http://erlang.org/doc/man/app.html), which is a file called
+  file*](https://www.erlang.org/doc/man/application.html), which is a file called
   `APP_NAME.app`. For example, the application resource file of the OTP
   application `ex_unit` is called `ex_unit.app`.
 
@@ -267,11 +293,10 @@ defmodule Application do
   ## Further information
 
   For further details on applications please check the documentation of the
-  [`application`](http://www.erlang.org/doc/man/application.html) Erlang module,
-  and the
-  [Applications](http://www.erlang.org/doc/design_principles/applications.html)
+  [`:application` Erlang module](`:application`), and the
+  [Applications](https://www.erlang.org/doc/design_principles/applications.html)
   section of the [OTP Design Principles User's
-  Guide](http://erlang.org/doc/design_principles/users_guide.html).
+  Guide](https://www.erlang.org/doc/design_principles/users_guide.html).
   """
 
   @doc """
@@ -386,6 +411,7 @@ defmodule Application do
     :maxT,
     :registered,
     :included_applications,
+    :optional_applications,
     :applications,
     :mod,
     :start_phases
@@ -499,21 +525,34 @@ defmodule Application do
   Giving a path is useful to let Elixir know that only certain paths
   in a large configuration are compile time dependent.
   """
-  # TODO: Warn on v1.14 if get_env/fetch_env/fetch_env! is used at compile time instead of compile_env
   @doc since: "1.10.0"
   @spec compile_env(app, key | list, value) :: value
-  defmacro compile_env(app, key_or_path, default \\ nil) when is_atom(app) do
+  defmacro compile_env(app, key_or_path, default \\ nil) do
     if __CALLER__.function do
       raise "Application.compile_env/3 cannot be called inside functions, only in the module body"
     end
 
+    key_or_path = Macro.expand_literals(key_or_path, %{__CALLER__ | function: {:__info__, 1}})
+
     quote do
-      Application.__compile_env__(unquote(app), unquote(key_or_path), unquote(default), __ENV__)
+      Application.compile_env(__ENV__, unquote(app), unquote(key_or_path), unquote(default))
     end
   end
 
-  @doc false
-  def __compile_env__(app, key_or_path, default, env) do
+  @doc """
+  Reads the application environment at compilation time from a macro.
+
+  Typically, developers will use `compile_env/3`. This function must
+  only be invoked from macros which aim to read the compilation environment
+  dynamically.
+
+  It expects a `Macro.Env` as first argument, where the `Macro.Env` is
+  typically the `__CALLER__` in a macro. It raises if `Macro.Env` comes
+  from a function.
+  """
+  @doc since: "1.14.0"
+  @spec compile_env(Macro.Env.t(), app, key | list, value) :: value
+  def compile_env(%Macro.Env{} = env, app, key_or_path, default) do
     case fetch_compile_env(app, key_or_path, env) do
       {:ok, value} -> value
       :error -> default
@@ -528,18 +567,33 @@ defmodule Application do
   """
   @doc since: "1.10.0"
   @spec compile_env!(app, key | list) :: value
-  defmacro compile_env!(app, key_or_path) when is_atom(app) do
+  defmacro compile_env!(app, key_or_path) do
     if __CALLER__.function do
       raise "Application.compile_env!/2 cannot be called inside functions, only in the module body"
     end
 
+    key_or_path = Macro.expand_literals(key_or_path, %{__CALLER__ | function: {:__info__, 1}})
+
     quote do
-      Application.__compile_env__!(unquote(app), unquote(key_or_path), __ENV__)
+      Application.compile_env!(__ENV__, unquote(app), unquote(key_or_path))
     end
   end
 
-  @doc false
-  def __compile_env__!(app, key_or_path, env) do
+  @doc """
+  Reads the application environment at compilation time from a macro
+  or raises.
+
+  Typically, developers will use `compile_env!/2`. This function must
+  only be invoked from macros which aim to read the compilation environment
+  dynamically.
+
+  It expects a `Macro.Env` as first argument, where the `Macro.Env` is
+  typically the `__CALLER__` in a macro. It raises if `Macro.Env` comes
+  from a function.
+  """
+  @doc since: "1.14.0"
+  @spec compile_env!(Macro.Env.t(), app, key | list) :: value
+  def compile_env!(%Macro.Env{} = env, app, key_or_path) do
     case fetch_compile_env(app, key_or_path, env) do
       {:ok, value} ->
         value
@@ -551,8 +605,9 @@ defmodule Application do
     end
   end
 
-  defp fetch_compile_env(app, key, env) when is_atom(key),
-    do: fetch_compile_env(app, key, [], env)
+  defp fetch_compile_env(app, key, env) when is_atom(key) do
+    fetch_compile_env(app, key, [], env)
+  end
 
   defp fetch_compile_env(app, [key | paths], env) when is_atom(key),
     do: fetch_compile_env(app, key, paths, env)
@@ -577,14 +632,13 @@ defmodule Application do
   If the configuration parameter does not exist, the function returns the
   `default` value.
 
-  **Important:** if you are reading the application environment at compilation
-  time, for example, inside the module definition instead of inside of a
-  function, see `compile_env/3` instead.
+  > **Important:** you must use this function to read only your own application
+  > environment. Do not read the environment of other applications.
 
-  **Important:** if you are writing a library to be used by other developers,
-  it is generally recommended to avoid the application environment, as the
-  application environment is effectively a global storage. For more information,
-  read our [library guidelines](library-guidelines.md).
+  > **Important:** if you are writing a library to be used by other developers,
+  > it is generally recommended to avoid the application environment, as the
+  > application environment is effectively a global storage. For more information,
+  > read our [library guidelines](library-guidelines.md).
 
   ## Examples
 
@@ -611,10 +665,18 @@ defmodule Application do
 
   Our database engine used by `:my_app` needs to know what databases exist, and
   what the database configurations are. The database engine can make a call to
-  `get_env(:my_app, :my_app_databases)` to retrieve the list of databases (specified
-  by module names). Our database engine can then traverse each repository in the
-  list and then call `get_env(:my_app, Databases.RepoOne)` and so forth to retrieve
-  the configuration of each one.
+  `Application.get_env(:my_app, :my_app_databases, [])` to retrieve the list of
+  databases (specified by module names).
+
+  The engine can then traverse each repository in the list and call
+  `Application.get_env(:my_app, Databases.RepoOne)` and so forth to retrieve the
+  configuration of each one. In this case, each configuration will be a keyword
+  list, so you can use the functions in the `Keyword` module or even the `Access`
+  module to traverse it, for example:
+
+      config = Application.get_env(:my_app, Databases.RepoOne)
+      config[:ip]
+
   """
   @spec get_env(app, key, value) :: value
   def get_env(app, key, default \\ nil) when is_atom(app) do
@@ -626,6 +688,14 @@ defmodule Application do
   Returns the value for `key` in `app`'s environment in a tuple.
 
   If the configuration parameter does not exist, the function returns `:error`.
+
+  > **Important:** you must use this function to read only your own application
+  > environment. Do not read the environment of other applications.
+
+  > **Important:** if you are writing a library to be used by other developers,
+  > it is generally recommended to avoid the application environment, as the
+  > application environment is effectively a global storage. For more information,
+  > read our [library guidelines](library-guidelines.md).
   """
   @spec fetch_env(app, key) :: {:ok, value} | :error
   def fetch_env(app, key) when is_atom(app) do
@@ -642,9 +712,13 @@ defmodule Application do
 
   If the configuration parameter does not exist, raises `ArgumentError`.
 
-  **Important:** if you are reading the application environment at compilation
-  time, for example, inside the module definition instead of inside of a
-  function, see `compile_env!/2` instead.
+  > **Important:** you must use this function to read only your own application
+  > environment. Do not read the environment of other applications.
+
+  > **Important:** if you are writing a library to be used by other developers,
+  > it is generally recommended to avoid the application environment, as the
+  > application environment is effectively a global storage. For more information,
+  > read our [library guidelines](library-guidelines.md).
   """
   @spec fetch_env!(app, key) :: value
   def fetch_env!(app, key) when is_atom(app) do
@@ -694,9 +768,6 @@ defmodule Application do
     :application.set_env(app, key, value, opts)
   end
 
-  # TODO: Remove this once we support Erlang/OTP 22+ exclusively.
-  @compile {:no_warn_undefined, {:application, :set_env, 2}}
-
   @doc """
   Puts the environment for multiple apps at the same time.
 
@@ -705,28 +776,14 @@ defmodule Application do
     * have the same application listed more than once
     * have the same key inside the same application listed more than once
 
-  If those conditions are not met, the behaviour is undefined
-  (on Erlang/OTP 21 and earlier) or will raise (on Erlang/OTP 22
-  and later).
+  If those conditions are not met, it will raise.
 
   It receives the same options as `put_env/4`. Returns `:ok`.
   """
   @doc since: "1.9.0"
   @spec put_all_env([{app, [{key, value}]}], timeout: timeout, persistent: boolean) :: :ok
   def put_all_env(config, opts \\ []) when is_list(config) and is_list(opts) do
-    # TODO: Remove function exported? check when we require Erlang/OTP 22+
-    if function_exported?(:application, :set_env, 2) do
-      :application.set_env(config, opts)
-    else
-      for app_keyword <- config,
-          {app, keyword} = app_keyword,
-          key_value <- keyword,
-          {key, value} = key_value do
-        :application.set_env(app, key, value, opts)
-      end
-
-      :ok
-    end
+    :application.set_env(config, opts)
   end
 
   @doc """
@@ -743,6 +800,7 @@ defmodule Application do
   defp maybe_warn_on_app_env_key(_app, key) when is_atom(key),
     do: :ok
 
+  # TODO: Remove this deprecation warning on 2.0+ and allow list lookups as in compile_env.
   defp maybe_warn_on_app_env_key(app, key) do
     message = "passing non-atom as application env key is deprecated, got: #{inspect(key)}"
     IO.warn_once({Application, :key, app, key}, message, _stacktrace_drop_levels = 2)
@@ -766,7 +824,7 @@ defmodule Application do
   @doc """
   Ensures the given `app` is loaded.
 
-  Same as `load/2` but returns `:ok` if the application was already
+  Same as `load/1` but returns `:ok` if the application was already
   loaded.
   """
   @doc since: "1.10.0"
@@ -887,7 +945,7 @@ defmodule Application do
       #=> "bar-123"
 
   For more information on code paths, check the `Code` module in
-  Elixir and also Erlang's [`:code` module](http://www.erlang.org/doc/man/code.html).
+  Elixir and also Erlang's [`:code` module](`:code`).
   """
   @spec app_dir(app) :: String.t()
   def app_dir(app) when is_atom(app) do
