@@ -57,7 +57,7 @@ defmodule Calendar do
   representing the microseconds to external format. If the precision is 0,
   it means microseconds must be skipped.
   """
-  @type microsecond :: {non_neg_integer, non_neg_integer}
+  @type microsecond :: {value :: non_neg_integer, precision :: non_neg_integer}
 
   @typedoc "A calendar implementation"
   @type calendar :: module
@@ -189,7 +189,7 @@ defmodule Calendar do
   @doc """
   Calculates the year and era from the given `year`.
   """
-  @callback year_of_era(year) :: {year, era}
+  @callback year_of_era(year, month, day) :: {year, era}
 
   @doc """
   Calculates the day and era from the given `year`, `month`, and `day`.
@@ -365,7 +365,7 @@ defmodule Calendar do
   """
   @doc since: "1.8.0"
   @spec put_time_zone_database(time_zone_database()) :: :ok
-  def put_time_zone_database(database) do
+  def put_time_zone_database(database) when is_atom(database) do
     Application.put_env(:elixir, :time_zone_database, database)
   end
 
@@ -375,7 +375,7 @@ defmodule Calendar do
   @doc since: "1.8.0"
   @spec get_time_zone_database() :: time_zone_database()
   def get_time_zone_database() do
-    Application.get_env(:elixir, :time_zone_database, Calendar.UTCOnlyTimeZoneDatabase)
+    Application.fetch_env!(:elixir, :time_zone_database)
   end
 
   @doc """
@@ -453,7 +453,7 @@ defmodule Calendar do
   b      | Abbreviated month name                                                  | Jan
   B      | Full month name                                                         | January
   c      | Preferred date+time representation                                      | 2018-10-17 12:34:56
-  d      | Day of the month                                                        | 01, 12
+  d      | Day of the month                                                        | 01, 31
   f      | Microseconds *(does not support width and padding modifiers)*           | 000000, 999999, 0123
   H      | Hour using a 24-hour clock                                              | 00, 23
   I      | Hour using a 12-hour clock                                              | 01, 12
@@ -485,6 +485,9 @@ defmodule Calendar do
       iex> Calendar.strftime(~U[2019-08-26 13:52:06.0Z], "%a, %B %d %Y")
       "Mon, August 26 2019"
 
+      iex> Calendar.strftime(~U[2020-04-02 13:52:06.0Z], "%B %-d, %Y")
+      "April 2, 2020"
+
       iex> Calendar.strftime(~U[2019-08-26 13:52:06.0Z], "%c")
       "2019-08-26 13:52:06"
 
@@ -541,8 +544,8 @@ defmodule Calendar do
     parse_modifiers(rest, width, "", parser_data)
   end
 
-  defp parse_modifiers("0" <> rest, width, nil, parser_data) do
-    parse_modifiers(rest, width, ?0, parser_data)
+  defp parse_modifiers("0" <> rest, nil, nil, parser_data) do
+    parse_modifiers(rest, nil, ?0, parser_data)
   end
 
   defp parse_modifiers("_" <> rest, width, nil, parser_data) do
@@ -788,7 +791,14 @@ defmodule Calendar do
 
   # Year
   defp format_modifiers("Y" <> rest, width, pad, datetime, format_options, acc) do
-    result = datetime.year |> Integer.to_string() |> pad_leading(width, pad)
+    {sign, year} =
+      if datetime.year < 0 do
+        {?-, -datetime.year}
+      else
+        {[], datetime.year}
+      end
+
+    result = [sign | year |> Integer.to_string() |> pad_leading(width, pad)]
     parse(rest, datetime, format_options, [result | acc])
   end
 
