@@ -8,16 +8,85 @@ defprotocol Inspect do
   The `Inspect` protocol converts an Elixir data structure into an
   algebra document.
 
+  This is typically done when you want to customize how your own
+  structs are inspected in logs and the terminal.
+
   This documentation refers to implementing the `Inspect` protocol
   for your own data structures. To learn more about using inspect,
   see `Kernel.inspect/2` and `IO.inspect/2`.
 
-  The `inspect/2` function receives the entity to be inspected
-  followed by the inspecting options, represented by the struct
-  `Inspect.Opts`. Building of the algebra document is done with
-  `Inspect.Algebra`.
+  ## Inspect representation
 
-  ## Examples
+  There are typically three choices of inspect representation. In order
+  to understand them, let's imagine we have the following `User` struct:
+
+      defmodule User do
+        defstruct [:id, :name, :address]
+      end
+
+  Our choices are:
+
+    1. Print the struct using Elixir's struct syntax, for example:
+       `%User{address: "Earth", id: 13, name: "Jane"}`. This is the
+       default representation and best choice if all struct fields
+       are public.
+
+    2. Print using the `#User<...>` notation, for example: `#User<id: 13, name: "Jane", ...>`.
+       This notation does not emit valid Elixir code and is typically
+       used when the struct has private fields (for example, you may want
+       to hide the field `:address` to redact person identifiable information).
+
+    3. Print the struct using the expression syntax, for example:
+       `User.new(13, "Jane", "Earth")`. This assumes there is a `User.new/3`
+       function. This option is mostly used as an alternative to option 2
+       for representing custom data structures, such as `MapSet`, `Date.Range`,
+       and others.
+
+  You can implement the Inspect protocol for your own structs while
+  adhering to the conventions above. Option 1 is the default representation
+  and you can quickly achieve option 2 by deriving the `Inspect` protocol.
+  For option 3, you need your custom implementation.
+
+  ## Deriving
+
+  The `Inspect` protocol can be derived to customize the order of fields
+  (the default is alphabetical) and hide certain fields from structs,
+  so they don't show up in logs, inspects and similar. The latter is
+  especially useful for fields containing private information.
+
+  The supported options are:
+
+    * `:only` - only include the given fields when inspecting.
+
+    * `:except` - remove the given fields when inspecting.
+
+    * `:optional` - (since v1.14.0) do not include a field if it
+      matches its default value. This can be used to simplify the
+      struct representation at the cost of hiding information.
+
+  Whenever `:only` or `:except` are used to restrict fields,
+  the struct will be printed using the `#User<...>` notation,
+  as the struct can no longer be copy and pasted as valid Elixir
+  code. Let's see an example:
+
+      defmodule User do
+        @derive {Inspect, only: [:id, :name]}
+        defstruct [:id, :name, :address]
+      end
+
+      inspect(%User{id: 1, name: "Jane", address: "Earth"})
+      #=> #User<id: 1, name: "Jane", ...>
+
+  If you use only the `:optional` option, the struct will still be
+  printed as `%User{...}`.
+
+  ## Custom implementation
+
+  You can also define your custom protocol implementation by
+  defining the `inspect/2` function. The function receives the
+  entity to be inspected followed by the inspecting options,
+  represented by the struct `Inspect.Opts`. Building of the
+  algebra document is done with `Inspect.Algebra`.
 
   Many times, inspecting a structure can be implemented in function
   of existing entities. For example, here is `MapSet`'s `inspect/2`
@@ -27,23 +96,24 @@ defprotocol Inspect do
         import Inspect.Algebra
 
         def inspect(map_set, opts) do
-          concat(["#MapSet<", to_doc(MapSet.to_list(map_set), opts), ">"])
+          concat(["MapSet.new(", Inspect.List.inspect(MapSet.to_list(map_set), opts), ")"])
         end
       end
 
   The [`concat/1`](`Inspect.Algebra.concat/1`) function comes from
   `Inspect.Algebra` and it concatenates algebra documents together.
-  In the example above it is concatenating the string `"#MapSet<"`,
+  In the example above it is concatenating the string `"MapSet.new("`,
   the document returned by `Inspect.Algebra.to_doc/2`, and the final
-  string `">"`. We prefix the module name `#` to denote the inspect
-  presentation is not actually valid Elixir syntax.
+  string `")"`. Therefore, the MapSet with the numbers 1, 2, and 3
+  will be printed as:
 
-  Finally, note strings themselves are valid algebra documents that
-  keep their formatting when pretty printed. This means your `Inspect`
-  implementation may simply return a string, although that will devoid
-  it of any pretty-printing.
+      iex> MapSet.new([1, 2, 3], fn x -> x * 2 end)
+      MapSet.new([2, 4, 6])
 
-  ## Error handling
+  In other words, `MapSet`'s inspect representation returns an expression
+  that, when evaluated, builds the `MapSet` itself.
+
+  ### Error handling
 
   In case there is an error while your structure is being inspected,
   Elixir will raise an `ArgumentError` error and will automatically fall back
@@ -55,24 +125,6 @@ defprotocol Inspect do
 
       Inspect.MapSet.inspect(MapSet.new(), %Inspect.Opts{})
 
-  ## Deriving
-
-  The `Inspect` protocol can be derived to hide certain fields from
-  structs, so they don't show up in logs, inspects and similar. This
-  is especially useful for fields containing private information.
-
-  The options `:only` and `:except` can be used with `@derive` to
-  specify which fields should and should not appear in the
-  algebra document:
-
-      defmodule User do
-        @derive {Inspect, only: [:id, :name]}
-        defstruct [:id, :name, :address]
-      end
-
-      inspect(%User{id: 1, name: "Homer", address: "742 Evergreen Terrace"})
-      #=> #User<id: 1, name: "Homer", ...>
-
   """
 
   # Handle structs in Any
@@ -94,7 +146,7 @@ defimpl Inspect, for: Atom do
   require Macro
 
   def inspect(atom, opts) do
-    color(Identifier.inspect_as_atom(atom), color_key(atom), opts)
+    color(Macro.inspect_atom(:literal, atom), color_key(atom), opts)
   end
 
   defp color_key(atom) when is_boolean(atom), do: :boolean
@@ -210,7 +262,7 @@ defimpl Inspect, for: List do
             {escaped, _} -> [?', escaped, ?', " ++ ..."]
           end
 
-        IO.iodata_to_binary(inspected)
+        color(IO.iodata_to_binary(inspected), :charlist, opts)
 
       keyword?(term) ->
         container_doc(open, term, close, opts, &keyword/2, separator: sep, break: :strict)
@@ -222,7 +274,7 @@ defimpl Inspect, for: List do
 
   @doc false
   def keyword({key, value}, opts) do
-    key = color(Identifier.inspect_as_key(key), :atom, opts)
+    key = color(Macro.inspect_atom(:key, key), :atom, opts)
     concat(key, concat(" ", to_doc(value, opts)))
   end
 
@@ -250,28 +302,38 @@ end
 
 defimpl Inspect, for: Map do
   def inspect(map, opts) do
-    inspect(map, "", opts)
+    list =
+      if Keyword.get(opts.custom_options, :sort_maps) do
+        map |> Map.to_list() |> :lists.sort()
+      else
+        Map.to_list(map)
+      end
+
+    fun =
+      if Inspect.List.keyword?(list) do
+        &Inspect.List.keyword/2
+      else
+        sep = color(" => ", :map, opts)
+        &to_assoc(&1, &2, sep)
+      end
+
+    map_container_doc(list, "", opts, fun)
   end
 
-  def inspect(map, name, opts) do
-    map = Map.to_list(map)
+  def inspect(map, name, infos, opts) do
+    fun = fn %{field: field}, opts -> Inspect.List.keyword({field, Map.get(map, field)}, opts) end
+    map_container_doc(infos, name, opts, fun)
+  end
+
+  defp to_assoc({key, value}, opts, sep) do
+    concat(concat(to_doc(key, opts), sep), to_doc(value, opts))
+  end
+
+  defp map_container_doc(list, name, opts, fun) do
     open = color("%" <> name <> "{", :map, opts)
     sep = color(",", :map, opts)
     close = color("}", :map, opts)
-    container_doc(open, map, close, opts, traverse_fun(map, opts), separator: sep, break: :strict)
-  end
-
-  defp traverse_fun(list, opts) do
-    if Inspect.List.keyword?(list) do
-      &Inspect.List.keyword/2
-    else
-      sep = color(" => ", :map, opts)
-      &to_map(&1, &2, sep)
-    end
-  end
-
-  defp to_map({key, value}, opts, sep) do
-    concat(concat(to_doc(key, opts), sep), to_doc(value, opts))
+    container_doc(open, list, close, opts, fun, separator: sep, break: :strict)
   end
 end
 
@@ -309,13 +371,31 @@ defimpl Inspect, for: Integer do
 end
 
 defimpl Inspect, for: Float do
-  def inspect(term, opts) do
-    inspected = IO.iodata_to_binary(:io_lib_format.fwrite_g(term))
-    color(inspected, :number, opts)
+  def inspect(float, opts) do
+    abs = abs(float)
+
+    formatted =
+      if abs >= 1.0 and abs < 1.0e16 and trunc(float) == float do
+        [Integer.to_string(trunc(float)), ?., ?0]
+      else
+        :io_lib_format.fwrite_g(float)
+      end
+
+    color(IO.iodata_to_binary(formatted), :number, opts)
   end
 end
 
 defimpl Inspect, for: Regex do
+  def inspect(regex = %{opts: regex_opts}, opts) when is_list(regex_opts) do
+    concat([
+      "Regex.compile!(",
+      Inspect.BitString.inspect(regex.source, opts),
+      ", ",
+      Inspect.List.inspect(regex_opts, opts),
+      ")"
+    ])
+  end
+
   def inspect(regex, opts) do
     {escaped, _} =
       regex.source
@@ -347,9 +427,12 @@ defimpl Inspect, for: Function do
     name = fun_info[:name]
 
     cond do
+      not is_atom(mod) ->
+        "#Function<#{uniq(fun_info)}/#{fun_info[:arity]}>"
+
       fun_info[:type] == :external and fun_info[:env] == [] ->
-        inspected_as_atom = Identifier.inspect_as_atom(mod)
-        inspected_as_function = Identifier.inspect_as_function(name)
+        inspected_as_atom = Macro.inspect_atom(:literal, mod)
+        inspected_as_function = Macro.inspect_atom(:remote_call, name)
         "&#{inspected_as_atom}.#{inspected_as_function}/#{fun_info[:arity]}"
 
       match?('elixir_compiler_' ++ _, Atom.to_charlist(mod)) ->
@@ -365,7 +448,7 @@ defimpl Inspect, for: Function do
   end
 
   defp default_inspect(mod, fun_info) do
-    inspected_as_atom = Identifier.inspect_as_atom(mod)
+    inspected_as_atom = Macro.inspect_atom(:literal, mod)
     extracted_name = extract_name(fun_info[:name])
     "#Function<#{uniq(fun_info)}/#{fun_info[:arity]} in #{inspected_as_atom}#{extracted_name}>"
   end
@@ -377,10 +460,10 @@ defimpl Inspect, for: Function do
   defp extract_name(name) do
     case Identifier.extract_anonymous_fun_parent(name) do
       {name, arity} ->
-        "." <> Identifier.inspect_as_function(name) <> "/" <> arity
+        "." <> Macro.inspect_atom(:remote_call, name) <> "/" <> arity
 
       :error ->
-        "." <> Identifier.inspect_as_function(name)
+        "." <> Macro.inspect_atom(:remote_call, name)
     end
   end
 
@@ -389,6 +472,36 @@ defimpl Inspect, for: Function do
   end
 end
 
+defimpl Inspect, for: Inspect.Error do
+  @impl true
+  def inspect(%{stacktrace: stacktrace} = inspect_error, _opts) do
+    message = Exception.message(inspect_error)
+    format_output(message, stacktrace)
+  end
+
+  defp format_output(message, [_ | _] = stacktrace) do
+    stacktrace = Exception.format_stacktrace(stacktrace)
+
+    """
+    #Inspect.Error<
+    #{Inspect.Error.pad(message, 2)}
+
+      Stacktrace:
+
+    #{stacktrace}
+    >\
+    """
+  end
+
+  defp format_output(message, []) do
+    """
+    #Inspect.Error<
+      #{Inspect.Error.pad(message, 2)}
+    >\
+    """
+  end
+end
+
 defimpl Inspect, for: PID do
   def inspect(pid, _opts) do
     "#PID" <> IO.iodata_to_binary(:erlang.pid_to_list(pid))
@@ -413,11 +526,11 @@ defimpl Inspect, for: Any do
     fields = Map.keys(struct) -- [:__exception__, :__struct__]
     only = Keyword.get(options, :only, fields)
     except = Keyword.get(options, :except, [])
+    optional = Keyword.get(options, :optional, [])
 
-    filtered_fields =
-      fields
-      |> Enum.reject(&(&1 in except))
-      |> Enum.filter(&(&1 in only))
+    :ok = validate_option(:only, only, fields, module)
+    :ok = validate_option(:except, except, fields, module)
+    :ok = validate_option(:optional, optional, fields, module)
 
     inspect_module =
       if fields == only and except == [] do
@@ -426,45 +539,86 @@ defimpl Inspect, for: Any do
         Inspect.Any
       end
 
+    filtered_fields =
+      fields
+      |> Enum.reject(&(&1 in except))
+      |> Enum.filter(&(&1 in only))
+
+    optional? =
+      if optional == [] do
+        false
+      else
+        optional_map = for field <- optional, into: %{}, do: {field, Map.fetch!(struct, field)}
+
+        quote do
+          case unquote(Macro.escape(optional_map)) do
+            %{^var!(field) => var!(default)} ->
+              var!(default) == Map.get(var!(struct), var!(field))
+
+            %{} ->
+              false
+          end
+        end
+      end
+
     quote do
       defimpl Inspect, for: unquote(module) do
         def inspect(var!(struct), var!(opts)) do
-          var!(map) = Map.take(var!(struct), unquote(filtered_fields))
-          var!(name) = Identifier.inspect_as_atom(unquote(module))
-          unquote(inspect_module).inspect(var!(map), var!(name), var!(opts))
+          var!(infos) =
+            for %{field: var!(field)} = var!(info) <- unquote(module).__info__(:struct),
+                var!(field) in unquote(filtered_fields) and not unquote(optional?),
+                do: var!(info)
+
+          var!(name) = Macro.inspect_atom(:literal, unquote(module))
+          unquote(inspect_module).inspect(var!(struct), var!(name), var!(infos), var!(opts))
         end
       end
     end
   end
 
+  defp validate_option(option, option_list, fields, module) do
+    case option_list -- fields do
+      [] ->
+        :ok
+
+      unknown_fields ->
+        raise ArgumentError,
+              "unknown fields #{Kernel.inspect(unknown_fields)} in #{Kernel.inspect(option)} " <>
+                "when deriving the Inspect protocol for #{Kernel.inspect(module)}"
+    end
+  end
+
   def inspect(%module{} = struct, opts) do
     try do
-      module.__struct__()
+      {module.__struct__(), module.__info__(:struct)}
     rescue
       _ -> Inspect.Map.inspect(struct, opts)
     else
-      dunder ->
+      {dunder, fields} ->
         if Map.keys(dunder) == Map.keys(struct) do
-          pruned = Map.drop(struct, [:__struct__, :__exception__])
-          Inspect.Map.inspect(pruned, Identifier.inspect_as_atom(module), opts)
+          infos =
+            for %{field: field} = info <- fields,
+                field not in [:__struct__, :__exception__],
+                do: info
+
+          Inspect.Map.inspect(struct, Macro.inspect_atom(:literal, module), infos, opts)
         else
           Inspect.Map.inspect(struct, opts)
         end
     end
   end
 
-  def inspect(map, name, opts) do
-    map = Map.to_list(map) ++ [:...]
+  def inspect(map, name, infos, opts) do
     open = color("#" <> name <> "<", :map, opts)
     sep = color(",", :map, opts)
     close = color(">", :map, opts)
 
     fun = fn
-      {key, value}, opts -> Inspect.List.keyword({key, value}, opts)
+      %{field: field}, opts -> Inspect.List.keyword({field, Map.get(map, field)}, opts)
       :..., _opts -> "..."
     end
 
-    container_doc(open, map, close, opts, fun, separator: sep, break: :strict)
+    container_doc(open, infos ++ [:...], close, opts, fun, separator: sep, break: :strict)
   end
 end
 
