@@ -31,8 +31,13 @@ defmodule Inspect.Opts do
       options. Useful when implementing the `Inspect` protocol for nested structs
       to pass the custom options through.
 
+      It supports some pre-defined keys:
+      - `:sort_maps` (since v1.15.0) - if set to `true`, sorts key-value pairs in maps.
+        This can be helpful to make map inspection deterministic for testing,
+        especially since key order is random since OTP 26.
+
     * `:inspect_fun` (since v1.9.0) - a function to build algebra documents.
-      Defaults to `Inspect.inspect/2`.
+      Defaults to `Inspect.Opts.default_inspect_fun/0`.
 
     * `:limit` - limits the number of items that are inspected for tuples,
       bitstrings, maps, lists and any other collection of items, with the exception of
@@ -53,7 +58,7 @@ defmodule Inspect.Opts do
     * `:safe` - when `false`, failures while inspecting structs will be raised
       as errors instead of being wrapped in the `Inspect.Error` exception. This
       is useful when debugging failures and crashes for custom inspect
-      implementations.
+      implementations. Defaults to `true`.
 
     * `:structs` - when `false`, structs are not formatted by the inspect
       protocol, they are instead printed as maps. Defaults to `true`.
@@ -64,6 +69,7 @@ defmodule Inspect.Opts do
       `:atom`, `:binary`, `:boolean`, `:list`, `:map`, `:number`, `:regex`,
       `:string`, and `:tuple`. Custom data types may provide their own options.
       Colors can be any `t:IO.ANSI.ansidata/0` as accepted by `IO.ANSI.format/1`.
+      A default list of colors can be retrieved from `IO.ANSI.syntax_colors/0`.
 
     * `:width` - number of characters per line used when pretty is `true` or when
       printing to IO devices. Set to `0` to force each item to be printed on its
@@ -89,11 +95,9 @@ defmodule Inspect.Opts do
 
   @type color_key :: atom
 
-  # TODO: Remove :char_lists key and :as_char_lists value on v2.0
   @type t :: %__MODULE__{
           base: :decimal | :binary | :hex | :octal,
           binaries: :infer | :as_binaries | :as_strings,
-          char_lists: :infer | :as_lists | :as_char_lists,
           charlists: :infer | :as_lists | :as_charlists,
           custom_options: keyword,
           inspect_fun: (any, t -> Inspect.Algebra.t()),
@@ -105,13 +109,62 @@ defmodule Inspect.Opts do
           syntax_colors: [{color_key, IO.ANSI.ansidata()}],
           width: non_neg_integer | :infinity
         }
-end
 
-defmodule Inspect.Error do
-  @moduledoc """
-  Raised when a struct cannot be inspected.
+  @doc """
+  Builds an `Inspect.Opts` struct.
   """
-  defexception [:message]
+  @doc since: "1.13.0"
+  @spec new(keyword()) :: t
+  def new(opts) do
+    struct(%Inspect.Opts{inspect_fun: default_inspect_fun()}, opts)
+  end
+
+  @doc """
+  Returns the default inspect function.
+  """
+  @doc since: "1.13.0"
+  @spec default_inspect_fun() :: (term, t -> Inspect.Algebra.t())
+  def default_inspect_fun do
+    :persistent_term.get({__MODULE__, :inspect_fun}, &Inspect.inspect/2)
+  end
+
+  @doc """
+  Sets the default inspect function.
+
+  Set this option with care as it will change how all values
+  in the system are inspected. The main use of this functionality
+  is to provide an entry point to filter inspected values,
+  in order for entities to comply with rules and legislations
+  on data security and data privacy.
+
+  It is **extremely discouraged** for libraries to set their own
+  function as this must be controlled by applications. Libraries
+  should instead define their own structs with custom inspect
+  implementations. If a library must change the default inspect
+  function, then it is best to define to ask users of your library
+  to explicitly call `default_inspect_fun/1` with your function of
+  choice.
+
+  The default is `Inspect.inspect/2`.
+
+  ## Examples
+
+      previous_fun = Inspect.Opts.default_inspect_fun()
+
+      Inspect.Opts.default_inspect_fun(fn
+        %{address: _} = map, opts ->
+          previous_fun.(%{map | address: "[REDACTED]"}, opts)
+
+        value, opts ->
+          previous_fun.(value, opts)
+      end)
+
+  """
+  @doc since: "1.13.0"
+  @spec default_inspect_fun((term, t -> Inspect.Algebra.t())) :: :ok
+  def default_inspect_fun(fun) when is_function(fun, 2) do
+    :persistent_term.put({__MODULE__, :inspect_fun}, fun)
+  end
 end
 
 defmodule Inspect.Algebra do
@@ -206,12 +259,18 @@ defmodule Inspect.Algebra do
           | doc_group
           | doc_nest
           | doc_string
+          | doc_limit
 
   @typep doc_string :: {:doc_string, t, non_neg_integer}
   defmacrop doc_string(string, length) do
     quote do: {:doc_string, unquote(string), unquote(length)}
   end
 
+  @typep doc_limit :: {:doc_limit, t, pos_integer | :infinity}
+  defmacrop doc_limit(doc, limit) do
+    quote do: {:doc_limit, unquote(doc), unquote(limit)}
+  end
+
   @typep doc_cons :: {:doc_cons, t, t}
   defmacrop doc_cons(left, right) do
     quote do: {:doc_cons, unquote(left), unquote(right)}
@@ -261,7 +320,8 @@ defmodule Inspect.Algebra do
     :doc_force,
     :doc_group,
     :doc_nest,
-    :doc_string
+    :doc_string,
+    :doc_limit
   ]
 
   defguard is_doc(doc)
@@ -298,19 +358,28 @@ defmodule Inspect.Algebra do
             try do
               Process.put(:inspect_trap, true)
 
-              res = Inspect.Map.inspect(struct, %{opts | syntax_colors: []})
-              res = IO.iodata_to_binary(format(res, :infinity))
+              inspected_struct =
+                struct
+                |> Inspect.Map.inspect(%{
+                  opts
+                  | syntax_colors: [],
+                    inspect_fun: Inspect.Opts.default_inspect_fun()
+                })
+                |> format(opts.width)
+                |> IO.iodata_to_binary()
 
-              message =
-                "got #{inspect(caught_exception.__struct__)} with message " <>
-                  "#{inspect(Exception.message(caught_exception))} while inspecting #{res}"
-
-              exception = Inspect.Error.exception(message: message)
+              inspect_error =
+                Inspect.Error.exception(
+                  exception: caught_exception,
+                  stacktrace: __STACKTRACE__,
+                  inspected_struct: inspected_struct
+                )
 
               if opts.safe do
-                Inspect.inspect(exception, opts)
+                opts = %{opts | inspect_fun: Inspect.Opts.default_inspect_fun()}
+                Inspect.inspect(inspect_error, opts)
               else
-                reraise(exception, __STACKTRACE__)
+                reraise(inspect_error, __STACKTRACE__)
               end
             after
               Process.delete(:inspect_trap)
@@ -404,8 +473,9 @@ defmodule Inspect.Algebra do
 
   defp container_each([term | terms], limit, opts, fun, acc, simple?)
        when is_list(terms) and is_limit(limit) do
-    limit = decrement(limit)
-    doc = fun.(term, %{opts | limit: limit})
+    new_limit = decrement(limit)
+    doc = fun.(term, %{opts | limit: new_limit})
+    limit = if doc == :doc_nil, do: limit, else: new_limit
     container_each(terms, limit, opts, fun, [doc | acc], simple? and simple?(doc))
   end
 
@@ -517,6 +587,10 @@ defmodule Inspect.Algebra do
     doc_cons(doc1, doc2)
   end
 
+  def no_limit(doc) do
+    doc_limit(doc, :infinity)
+  end
+
   @doc ~S"""
   Concatenates a list of documents returning a new document.
 
@@ -536,7 +610,7 @@ defmodule Inspect.Algebra do
   Colors a document if the `color_key` has a color in the options.
   """
   @doc since: "1.4.0"
-  @spec color(t, Inspect.Opts.color_key(), Inspect.Opts.t()) :: doc_color
+  @spec color(t, Inspect.Opts.color_key(), Inspect.Opts.t()) :: t
   def color(doc, color_key, %Inspect.Opts{syntax_colors: syntax_colors}) when is_doc(doc) do
     if precolor = Keyword.get(syntax_colors, color_key) do
       postcolor = Keyword.get(syntax_colors, :reset, :reset)
@@ -566,7 +640,7 @@ defmodule Inspect.Algebra do
       ["hello", "\n     ", "world"]
 
   """
-  @spec nest(t, non_neg_integer | :cursor | :reset, :always | :break) :: doc_nest
+  @spec nest(t, non_neg_integer | :cursor | :reset, :always | :break) :: doc_nest | t
   def nest(doc, level, mode \\ :always)
 
   def nest(doc, :cursor, mode) when is_doc(doc) and mode in [:always, :break] do
@@ -894,12 +968,12 @@ defmodule Inspect.Algebra do
     format(width, 0, [{0, :flat, doc}])
   end
 
-  # Type representing the document mode to be rendered
+  # Type representing the document mode to be rendered:
   #
   #   * flat - represents a document with breaks as flats (a break may fit, as it may break)
   #   * break - represents a document with breaks as breaks (a break always fits, since it breaks)
   #
-  # The following modes are exclusive to fitting
+  # The following modes are exclusive to fitting:
   #
   #   * flat_no_break - represents a document with breaks as flat not allowed to enter in break mode
   #   * break_no_flat - represents a document with breaks as breaks not allowed to enter in flat mode
@@ -907,7 +981,7 @@ defmodule Inspect.Algebra do
   @typep mode :: :flat | :flat_no_break | :break | :break_no_flat
 
   @spec fits?(
-          width :: non_neg_integer(),
+          width :: non_neg_integer() | :infinity,
           column :: non_neg_integer(),
           break? :: boolean(),
           entries
@@ -972,9 +1046,17 @@ defmodule Inspect.Algebra do
   defp fits?(w, k, b?, [{i, m, doc_group(x, _)} | t]),
     do: fits?(w, k, b?, [{i, m, x} | {:tail, b?, t}])
 
-  @spec format(width :: non_neg_integer() | :infinity, column :: non_neg_integer(), [
-          {integer, mode, t}
-        ]) :: [binary]
+  defp fits?(w, k, b?, [{i, m, doc_limit(x, :infinity)} | t]) when w != :infinity,
+    do: fits?(:infinity, k, b?, [{i, :flat, x}, {i, m, doc_limit(empty(), w)} | t])
+
+  defp fits?(_w, k, b?, [{i, m, doc_limit(x, w)} | t]),
+    do: fits?(w, k, b?, [{i, m, x} | t])
+
+  @spec format(
+          width :: non_neg_integer() | :infinity,
+          column :: non_neg_integer(),
+          [{integer, mode, t}]
+        ) :: [binary]
   defp format(_, _, []), do: []
   defp format(w, k, [{_, _, :doc_nil} | t]), do: format(w, k, t)
   defp format(w, _, [{i, _, :doc_line} | t]), do: [indent(i) | format(w, i, t)]
@@ -1028,6 +1110,15 @@ defmodule Inspect.Algebra do
     end
   end
 
+  # Limit is set to infinity and then reverts
+  defp format(w, k, [{i, m, doc_limit(x, :infinity)} | t]) when w != :infinity do
+    format(:infinity, k, [{i, :flat, x}, {i, m, doc_limit(empty(), w)} | t])
+  end
+
+  defp format(_w, k, [{i, m, doc_limit(x, w)} | t]) do
+    format(w, k, [{i, m, x} | t])
+  end
+
   defp collapse(["\n" <> _ | t], max, count, i) do
     collapse(t, max, count + 1, i)
   end
