@@ -1,7 +1,8 @@
 defmodule IO.ANSI.Docs do
   @moduledoc false
 
-  @bullet_text "• "
+  @bullet_text_unicode "• "
+  @bullet_text_ascii "* "
   @bullets [?*, ?-, ?+]
   @spaces [" ", "\n", "\t"]
 
@@ -77,7 +78,9 @@ defmodule IO.ANSI.Docs do
   @metadata_filter [:deprecated, :guard, :since]
 
   defp print_each_metadata(metadata, options) do
-    Enum.reduce(metadata, false, fn
+    metadata
+    |> Enum.sort()
+    |> Enum.reduce(false, fn
       {key, value}, _printed when is_binary(value) and key in @metadata_filter ->
         label = metadata_label(key, options)
         indent = String.duplicate(" ", length_without_escape(label, 0) + 1)
@@ -193,6 +196,10 @@ defmodule IO.ANSI.Docs do
     inline_text("*", traverse_erlang_html(entries, indent, options), options)
   end
 
+  defp traverse_erlang_html({tag, _, entries}, indent, options) when tag in [:strong, :b] do
+    inline_text("**", traverse_erlang_html(entries, indent, options), options)
+  end
+
   defp traverse_erlang_html({:code, _, entries}, indent, options) do
     inline_text("`", traverse_erlang_html(entries, indent, options), options)
   end
@@ -219,7 +226,10 @@ defmodule IO.ANSI.Docs do
   end
 
   defp traverse_erlang_html({:dt, _, entries}, indent, options) do
-    ["#{indent}  ", @bullet_text | handle_erlang_html_text(entries, indent <> "    ", options)]
+    [
+      "#{indent}  ",
+      bullet_text(options) | handle_erlang_html_text(entries, indent <> "    ", options)
+    ]
   end
 
   defp traverse_erlang_html({:dd, _, entries}, indent, options) do
@@ -240,7 +250,10 @@ defmodule IO.ANSI.Docs do
       end
     else
       for {:li, _, lines} <- entries do
-        ["#{indent}  ", @bullet_text | handle_erlang_html_text(lines, indent <> "    ", options)]
+        [
+          "#{indent}  ",
+          bullet_text(options) | handle_erlang_html_text(lines, indent <> "    ", options)
+        ]
       end
     end
   end
@@ -288,7 +301,7 @@ defmodule IO.ANSI.Docs do
   end
 
   defp inline_html?(binary) when is_binary(binary), do: true
-  defp inline_html?({tag, _, _}) when tag in [:a, :code, :em, :i, :br], do: true
+  defp inline_html?({tag, _, _}) when tag in [:a, :code, :em, :i, :strong, :b, :br], do: true
   defp inline_html?(_), do: false
 
   ## Markdown
@@ -424,7 +437,7 @@ defmodule IO.ANSI.Docs do
     case stripped do
       <<bullet, ?\s, item::binary>> when bullet in @bullets ->
         write_text(text, indent, options)
-        process_list(@bullet_text, item, rest, count, indent, options)
+        process_list(bullet_text(options), item, rest, count, indent, options)
 
       <<d1, ?., ?\s, item::binary>> when d1 in ?0..?9 ->
         write_text(text, indent, options)
@@ -791,14 +804,14 @@ defmodule IO.ANSI.Docs do
     Regex.replace(~r{\[([^\]]*?)\]\((.*?)\)}, text, "\\1 (\\2)")
   end
 
-  # We have four entries: **, *, _ and `.
+  # We have four entries: **, __, *, _ and `.
   #
-  # The first three behave the same while the last one is simpler
+  # The first four behave the same while the last one is simpler
   # when it comes to delimiters as it ignores spaces and escape
-  # characters. But, since the first has two characters, we need to
-  # handle 3 cases:
+  # characters. But, since the first two has two characters,
+  # we need to handle 3 cases:
   #
-  # 1. **
+  # 1. __ and **
   # 2. _ and *
   # 3. `
   #
@@ -812,8 +825,8 @@ defmodule IO.ANSI.Docs do
 
   ### Inline start
 
-  defp handle_inline(<<?*, ?*, rest::binary>>, options) do
-    handle_inline(rest, ?d, ["**"], [], options)
+  defp handle_inline(<<mark, mark, rest::binary>>, options) when mark in @single do
+    handle_inline(rest, [mark | mark], [<<mark, mark>>], [], options)
   end
 
   defp handle_inline(<<mark, rest::binary>>, options) when mark in @single do
@@ -826,9 +839,10 @@ defmodule IO.ANSI.Docs do
 
   ### Inline delimiters
 
-  defp handle_inline(<<delimiter, ?*, ?*, rest::binary>>, nil, buffer, acc, options)
-       when rest != "" and delimiter in @delimiters do
-    handle_inline(rest, ?d, ["**"], [delimiter, Enum.reverse(buffer) | acc], options)
+  defp handle_inline(<<delimiter, mark, mark, rest::binary>>, nil, buffer, acc, options)
+       when rest != "" and delimiter in @delimiters and mark in @single do
+    acc = [delimiter, Enum.reverse(buffer) | acc]
+    handle_inline(rest, [mark | mark], [<<mark, mark>>], acc, options)
   end
 
   defp handle_inline(<<delimiter, mark, rest::binary>>, nil, buffer, acc, options)
@@ -843,9 +857,10 @@ defmodule IO.ANSI.Docs do
 
   ### Clauses for handling escape
 
-  defp handle_inline(<<?\\, ?\\, ?*, ?*, rest::binary>>, nil, buffer, acc, options)
-       when rest != "" do
-    handle_inline(rest, ?d, ["**"], [?\\, Enum.reverse(buffer) | acc], options)
+  defp handle_inline(<<?\\, ?\\, mark, mark, rest::binary>>, nil, buffer, acc, options)
+       when rest != "" and mark in @single do
+    acc = [?\\, Enum.reverse(buffer) | acc]
+    handle_inline(rest, [mark | mark], [<<mark, mark>>], acc, options)
   end
 
   defp handle_inline(<<?\\, ?\\, mark, rest::binary>>, nil, buffer, acc, options)
@@ -864,8 +879,8 @@ defmodule IO.ANSI.Docs do
 
   ### Inline end
 
-  defp handle_inline(<<?*, ?*, delimiter, rest::binary>>, ?d, buffer, acc, options)
-       when delimiter in @delimiters do
+  defp handle_inline(<<mark, mark, delimiter, rest::binary>>, [mark | mark], buffer, acc, options)
+       when delimiter in @delimiters and mark in @single do
     inline_buffer = inline_buffer(buffer, options)
     handle_inline(<<delimiter, rest::binary>>, nil, [], [inline_buffer | acc], options)
   end
@@ -876,8 +891,8 @@ defmodule IO.ANSI.Docs do
     handle_inline(<<delimiter, rest::binary>>, nil, [], [inline_buffer | acc], options)
   end
 
-  defp handle_inline(<<?*, ?*, rest::binary>>, ?d, buffer, acc, options)
-       when rest == "" do
+  defp handle_inline(<<mark, mark, rest::binary>>, [mark | mark], buffer, acc, options)
+       when rest == "" and mark in @single do
     handle_inline(<<>>, nil, [], [inline_buffer(buffer, options) | acc], options)
   end
 
@@ -923,13 +938,18 @@ defmodule IO.ANSI.Docs do
 
   defp color_for(mark, colors) do
     case mark do
-      "`" -> color(:doc_inline_code, colors)
-      "_" -> color(:doc_underline, colors)
-      "*" -> color(:doc_bold, colors)
+      "__" -> color(:doc_bold, colors)
       "**" -> color(:doc_bold, colors)
+      "_" -> color(:doc_underline, colors)
+      "*" -> color(:doc_underline, colors)
+      "`" -> color(:doc_inline_code, colors)
     end
   end
 
+  defp bullet_text(options) do
+    if options[:enabled], do: @bullet_text_unicode, else: @bullet_text_ascii
+  end
+
   defp color(style, colors) do
     IO.ANSI.format_fragment(colors[style], colors[:enabled])
   end
