@@ -186,6 +186,9 @@ defmodule Stream do
       iex> Stream.chunk_every([1, 2, 3, 4, 5, 6], 3, 3, []) |> Enum.to_list()
       [[1, 2, 3], [4, 5, 6]]
 
+      iex> Stream.chunk_every([1, 2, 3, 4], 3, 3, Stream.cycle([0])) |> Enum.to_list()
+      [[1, 2, 3], [4, 0, 0]]
+
   """
   @doc since: "1.5.0"
   @spec chunk_every(Enumerable.t(), pos_integer, pos_integer, Enumerable.t() | :discard) ::
@@ -417,10 +420,47 @@ defmodule Stream do
     lazy(enum, true, fn f1 -> R.drop_while(fun, f1) end)
   end
 
+  @doc """
+  Duplicates the given element `n` times in a stream.
+
+  `n` is an integer greater than or equal to `0`.
+
+  If `n` is `0`, an empty stream is returned.
+
+  ## Examples
+
+      iex> stream = Stream.duplicate("hello", 0)
+      iex> Enum.to_list(stream)
+      []
+
+      iex> stream = Stream.duplicate("hi", 1)
+      iex> Enum.to_list(stream)
+      ["hi"]
+
+      iex> stream = Stream.duplicate("bye", 2)
+      iex> Enum.to_list(stream)
+      ["bye", "bye"]
+
+      iex> stream = Stream.duplicate([1, 2], 3)
+      iex> Enum.to_list(stream)
+      [[1, 2], [1, 2], [1, 2]]
+  """
+  @doc since: "1.14.0"
+  @spec duplicate(any, non_neg_integer) :: Enumerable.t()
+  def duplicate(value, n) when is_integer(n) and n >= 0 do
+    unfold(n, fn
+      0 -> nil
+      remaining -> {value, remaining - 1}
+    end)
+  end
+
   @doc """
   Executes the given function for each element.
 
-  Useful for adding side effects (like printing) to a stream.
+  The values in the stream do not change, therefore this
+  function is useful for adding side effects (like printing)
+  to a stream. See `map/2` if producing a different stream
+  is desired.
 
   ## Examples
 
@@ -800,10 +840,10 @@ defmodule Stream do
   @doc """
   Transforms an existing stream.
 
-  It expects an accumulator and a function that receives each stream element
-  and an accumulator. It must return a tuple, where the first element is a new
-  stream (often a list) or the atom `:halt`, and the second element is the
-  accumulator to be used by the next element, if any, in both cases.
+  It expects an accumulator and a function that receives two arguments,
+  the stream element and the updated accumulator. It must return a tuple,
+  where the first element is a new stream (often a list) or the atom `:halt`,
+  and the second element is the accumulator to be used by the next element.
 
   Note: this function is equivalent to `Enum.flat_map_reduce/3`, except this
   function does not return the accumulator once the stream is processed.
@@ -822,44 +862,72 @@ defmodule Stream do
       iex> Enum.to_list(stream)
       [1001, 1002, 1003]
 
+  `Stream.transform/5` further generalizes this function to allow wrapping
+  around resources.
   """
   @spec transform(Enumerable.t(), acc, fun) :: Enumerable.t()
         when fun: (element, acc -> {Enumerable.t(), acc} | {:halt, acc}),
              acc: any
   def transform(enum, acc, reducer) when is_function(reducer, 2) do
-    &do_transform(enum, fn -> acc end, reducer, &1, &2, nil)
+    &do_transform(enum, fn -> acc end, reducer, &1, &2, nil, fn acc -> acc end)
   end
 
   @doc """
-  Transforms an existing stream with function-based start and finish.
-
-  The accumulator is only calculated when transformation starts. It also
-  allows an after function to be given which is invoked when the stream
-  halts or completes.
+  Similar to `Stream.transform/5`, except `last_fun` is not supplied.
 
   This function can be seen as a combination of `Stream.resource/3` with
   `Stream.transform/3`.
   """
-  @spec transform(Enumerable.t(), (() -> acc), fun, (acc -> term)) :: Enumerable.t()
-        when fun: (element, acc -> {Enumerable.t(), acc} | {:halt, acc}),
+  @spec transform(Enumerable.t(), start_fun, reducer, after_fun) :: Enumerable.t()
+        when start_fun: (() -> acc),
+             reducer: (element, acc -> {Enumerable.t(), acc} | {:halt, acc}),
+             after_fun: (acc -> term),
              acc: any
   def transform(enum, start_fun, reducer, after_fun)
       when is_function(start_fun, 0) and is_function(reducer, 2) and is_function(after_fun, 1) do
-    &do_transform(enum, start_fun, reducer, &1, &2, after_fun)
+    &do_transform(enum, start_fun, reducer, &1, &2, nil, after_fun)
   end
 
-  defp do_transform(enumerables, user_acc, user, inner_acc, fun, after_fun) do
+  @doc """
+  Transforms an existing stream with function-based start, last, and after
+  callbacks.
+
+  Once transformation starts, `start_fun` is invoked to compute the initial
+  accumulator. Then, for each element in the enumerable, the `reducer` function
+  is invoked with the element and the accumulator, returning new elements and a
+  new accumulator, as in `transform/3`.
+
+  Once the collection is done, `last_fun` is invoked with the accumulator to
+  emit any remaining items. Then `after_fun` is invoked, to close any resource,
+  but not emitting any new items. `last_fun` is only invoked if the given
+  enumerable terminates successfully (either because it is done or it halted
+  itself). `after_fun` is always invoked, therefore `after_fun` must be the
+  one used for closing resources.
+  """
+  @spec transform(Enumerable.t(), start_fun, reducer, last_fun, after_fun) :: Enumerable.t()
+        when start_fun: (() -> acc),
+             reducer: (element, acc -> {Enumerable.t(), acc} | {:halt, acc}),
+             last_fun: (acc -> {Enumerable.t(), acc} | {:halt, acc}),
+             after_fun: (acc -> term),
+             acc: any
+  def transform(enum, start_fun, reducer, last_fun, after_fun)
+      when is_function(start_fun, 0) and is_function(reducer, 2) and is_function(last_fun, 1) and
+             is_function(after_fun, 1) do
+    &do_transform(enum, start_fun, reducer, &1, &2, last_fun, after_fun)
+  end
+
+  defp do_transform(enumerables, user_acc, user, inner_acc, fun, last_fun, after_fun) do
     inner = &do_transform_each(&1, &2, fun)
     step = &do_transform_step(&1, &2)
     next = &Enumerable.reduce(enumerables, &1, step)
-    funs = {user, fun, inner, after_fun}
+    funs = {user, fun, inner, last_fun, after_fun}
     do_transform(user_acc.(), :cont, next, inner_acc, funs)
   end
 
   defp do_transform(user_acc, _next_op, next, {:halt, inner_acc}, funs) do
-    {_, _, _, after_fun} = funs
+    {_, _, _, _, after_fun} = funs
     next.({:halt, []})
-    do_after(after_fun, user_acc)
+    after_fun.(user_acc)
     {:halted, inner_acc}
   end
 
@@ -867,72 +935,99 @@ defmodule Stream do
     {:suspended, inner_acc, &do_transform(user_acc, next_op, next, &1, funs)}
   end
 
-  defp do_transform(user_acc, :halt, _next, {_, inner_acc}, funs) do
-    {_, _, _, after_fun} = funs
-    do_after(after_fun, user_acc)
-    {:halted, inner_acc}
-  end
-
   defp do_transform(user_acc, :cont, next, inner_acc, funs) do
-    {_, _, _, after_fun} = funs
+    {_, _, _, _, after_fun} = funs
 
     try do
       next.({:cont, []})
     catch
       kind, reason ->
-        do_after(after_fun, user_acc)
+        after_fun.(user_acc)
         :erlang.raise(kind, reason, __STACKTRACE__)
     else
       {:suspended, vals, next} ->
         do_transform_user(:lists.reverse(vals), user_acc, :cont, next, inner_acc, funs)
 
       {_, vals} ->
-        do_transform_user(:lists.reverse(vals), user_acc, :halt, next, inner_acc, funs)
+        do_transform_user(:lists.reverse(vals), user_acc, :last, next, inner_acc, funs)
     end
   end
 
+  defp do_transform(user_acc, :last, next, inner_acc, funs) do
+    {_, _, _, last_fun, after_fun} = funs
+
+    if last_fun do
+      try do
+        last_fun.(user_acc)
+      catch
+        kind, reason ->
+          next.({:halt, []})
+          after_fun.(user_acc)
+          :erlang.raise(kind, reason, __STACKTRACE__)
+      else
+        result -> do_transform_result(result, [], :halt, next, inner_acc, funs)
+      end
+    else
+      do_transform(user_acc, :halt, next, inner_acc, funs)
+    end
+  end
+
+  defp do_transform(user_acc, :halt, _next, inner_acc, funs) do
+    {_, _, _, _, after_fun} = funs
+    after_fun.(user_acc)
+    {:halted, elem(inner_acc, 1)}
+  end
+
   defp do_transform_user([], user_acc, next_op, next, inner_acc, funs) do
     do_transform(user_acc, next_op, next, inner_acc, funs)
   end
 
   defp do_transform_user([val | vals], user_acc, next_op, next, inner_acc, funs) do
-    {user, fun, inner, after_fun} = funs
+    {user, _, _, _, after_fun} = funs
 
     try do
       user.(val, user_acc)
     catch
       kind, reason ->
         next.({:halt, []})
-        do_after(after_fun, user_acc)
+        after_fun.(user_acc)
         :erlang.raise(kind, reason, __STACKTRACE__)
     else
+      result -> do_transform_result(result, vals, next_op, next, inner_acc, funs)
+    end
+  end
+
+  defp do_transform_result(result, vals, next_op, next, inner_acc, funs) do
+    {_, fun, inner, _, after_fun} = funs
+
+    case result do
       {[], user_acc} ->
         do_transform_user(vals, user_acc, next_op, next, inner_acc, funs)
 
       {list, user_acc} when is_list(list) ->
         reduce = &Enumerable.List.reduce(list, &1, fun)
-        do_list_transform(vals, user_acc, next_op, next, inner_acc, reduce, funs)
+        do_transform_inner_list(vals, user_acc, next_op, next, inner_acc, reduce, funs)
 
       {:halt, user_acc} ->
         next.({:halt, []})
-        do_after(after_fun, user_acc)
+        after_fun.(user_acc)
         {:halted, elem(inner_acc, 1)}
 
       {other, user_acc} ->
         reduce = &Enumerable.reduce(other, &1, inner)
-        do_enum_transform(vals, user_acc, next_op, next, inner_acc, reduce, funs)
+        do_transform_inner_enum(vals, user_acc, next_op, next, inner_acc, reduce, funs)
     end
   end
 
-  defp do_list_transform(vals, user_acc, next_op, next, inner_acc, reduce, funs) do
-    {_, _, _, after_fun} = funs
+  defp do_transform_inner_list(vals, user_acc, next_op, next, inner_acc, reduce, funs) do
+    {_, _, _, _, after_fun} = funs
 
     try do
       reduce.(inner_acc)
     catch
       kind, reason ->
         next.({:halt, []})
-        do_after(after_fun, user_acc)
+        after_fun.(user_acc)
         :erlang.raise(kind, reason, __STACKTRACE__)
     else
       {:done, acc} ->
@@ -940,24 +1035,24 @@ defmodule Stream do
 
       {:halted, acc} ->
         next.({:halt, []})
-        do_after(after_fun, user_acc)
+        after_fun.(user_acc)
         {:halted, acc}
 
       {:suspended, acc, continuation} ->
-        resume = &do_list_transform(vals, user_acc, next_op, next, &1, continuation, funs)
+        resume = &do_transform_inner_list(vals, user_acc, next_op, next, &1, continuation, funs)
         {:suspended, acc, resume}
     end
   end
 
-  defp do_enum_transform(vals, user_acc, next_op, next, {op, inner_acc}, reduce, funs) do
-    {_, _, _, after_fun} = funs
+  defp do_transform_inner_enum(vals, user_acc, next_op, next, {op, inner_acc}, reduce, funs) do
+    {_, _, _, _, after_fun} = funs
 
     try do
       reduce.({op, [:outer | inner_acc]})
     catch
       kind, reason ->
         next.({:halt, []})
-        do_after(after_fun, user_acc)
+        after_fun.(user_acc)
         :erlang.raise(kind, reason, __STACKTRACE__)
     else
       # Only take into account outer halts when the op is not halt itself.
@@ -967,21 +1062,18 @@ defmodule Stream do
 
       {:halted, [_ | acc]} ->
         next.({:halt, []})
-        do_after(after_fun, user_acc)
+        after_fun.(user_acc)
         {:halted, acc}
 
       {:done, [_ | acc]} ->
         do_transform_user(vals, user_acc, next_op, next, {:cont, acc}, funs)
 
       {:suspended, [_ | acc], continuation} ->
-        resume = &do_enum_transform(vals, user_acc, next_op, next, &1, continuation, funs)
+        resume = &do_transform_inner_enum(vals, user_acc, next_op, next, &1, continuation, funs)
         {:suspended, acc, resume}
     end
   end
 
-  defp do_after(nil, _user_acc), do: :ok
-  defp do_after(fun, user_acc), do: fun.(user_acc)
-
   defp do_transform_each(x, [:outer | acc], f) do
     case f.(x, acc) do
       {:halt, res} -> {:halt, [:inner | res]}
@@ -1105,9 +1197,9 @@ defmodule Stream do
   end
 
   @doc """
-  Zips two collections together, lazily.
+  Zips two enumerables together, lazily.
 
-  The zipping finishes as soon as any enumerable completes.
+  The zipping finishes as soon as either enumerable completes.
 
   ## Examples
 
@@ -1118,7 +1210,9 @@ defmodule Stream do
 
   """
   @spec zip(Enumerable.t(), Enumerable.t()) :: Enumerable.t()
-  def zip(left, right), do: zip([left, right])
+  def zip(enumerable1, enumerable2) do
+    zip_with(enumerable1, enumerable2, fn left, right -> {left, right} end)
+  end
 
   @doc """
   Zips corresponding elements from a finite collection of enumerables
@@ -1137,69 +1231,177 @@ defmodule Stream do
   @doc since: "1.4.0"
   @spec zip(enumerables) :: Enumerable.t() when enumerables: [Enumerable.t()] | Enumerable.t()
   def zip(enumerables) do
-    &prepare_zip(enumerables, &1, &2)
+    zip_with(enumerables, &List.to_tuple(&1))
   end
 
-  defp prepare_zip(enumerables, acc, fun) do
-    step = &do_zip_step(&1, &2)
+  @doc """
+  Lazily zips corresponding elements from two enumerables into a new one, transforming them with
+  the `zip_fun` function as it goes.
+
+  The `zip_fun` will be called with the first element from `enumerable1` and the first
+  element from `enumerable2`, then with the second element from each, and so on until
+  either one of the enumerables completes.
+
+  ## Examples
+
+      iex> concat = Stream.concat(1..3, 4..6)
+      iex> Stream.zip_with(concat, concat, fn a, b -> a + b end) |> Enum.to_list()
+      [2, 4, 6, 8, 10, 12]
+
+  """
+  @doc since: "1.12.0"
+  @spec zip_with(Enumerable.t(), Enumerable.t(), (term, term -> term)) :: Enumerable.t()
+  def zip_with(enumerable1, enumerable2, zip_fun)
+      when is_list(enumerable1) and is_list(enumerable2) and is_function(zip_fun, 2) do
+    &zip_pair(enumerable1, enumerable2, &1, &2, zip_fun)
+  end
+
+  def zip_with(enumerable1, enumerable2, zip_fun) when is_function(zip_fun, 2) do
+    zip_with([enumerable1, enumerable2], fn [left, right] -> zip_fun.(left, right) end)
+  end
+
+  defp zip_pair(_list1, _list2, {:halt, acc}, _fun, _zip_fun) do
+    {:halted, acc}
+  end
+
+  defp zip_pair(list1, list2, {:suspend, acc}, fun, zip_fun) do
+    {:suspended, acc, &zip_pair(list1, list2, &1, fun, zip_fun)}
+  end
+
+  defp zip_pair([], _list2, {:cont, acc}, _fun, _zip_fun), do: {:done, acc}
+  defp zip_pair(_list1, [], {:cont, acc}, _fun, _zip_fun), do: {:done, acc}
+
+  defp zip_pair([head1 | tail1], [head2 | tail2], {:cont, acc}, fun, zip_fun) do
+    zip_pair(tail1, tail2, fun.(zip_fun.(head1, head2), acc), fun, zip_fun)
+  end
+
+  @doc """
+  Lazily zips corresponding elements from a finite collection of enumerables into a new
+  enumerable, transforming them with the `zip_fun` function as it goes.
+
+  The first element from each of the enums in `enumerables` will be put into a list which is then passed to
+  the one-arity `zip_fun` function. Then, the second elements from each of the enums are put into a list and passed to
+  `zip_fun`, and so on until any one of the enums in `enumerables` completes.
+
+  Returns a new enumerable with the results of calling `zip_fun`.
+
+  ## Examples
+
+      iex> concat = Stream.concat(1..3, 4..6)
+      iex> Stream.zip_with([concat, concat], fn [a, b] -> a + b end) |> Enum.to_list()
+      [2, 4, 6, 8, 10, 12]
+
+      iex> concat = Stream.concat(1..3, 4..6)
+      iex> Stream.zip_with([concat, concat, 1..3], fn [a, b, c] -> a + b + c end) |> Enum.to_list()
+      [3, 6, 9]
+
+  """
+  @doc since: "1.12.0"
+  @spec zip_with(enumerables, (Enumerable.t() -> term)) :: Enumerable.t()
+        when enumerables: [Enumerable.t()] | Enumerable.t()
+  def zip_with(enumerables, zip_fun) when is_function(zip_fun, 1) do
+    if is_list(enumerables) and :lists.all(&is_list/1, enumerables) do
+      &zip_list(enumerables, &1, &2, zip_fun)
+    else
+      &zip_enum(enumerables, &1, &2, zip_fun)
+    end
+  end
+
+  defp zip_list(_enumerables, {:halt, acc}, _fun, _zip_fun) do
+    {:halted, acc}
+  end
+
+  defp zip_list(enumerables, {:suspend, acc}, fun, zip_fun) do
+    {:suspended, acc, &zip_list(enumerables, &1, fun, zip_fun)}
+  end
+
+  defp zip_list([], {:cont, acc}, _fun, _zip_fun), do: {:done, acc}
+
+  defp zip_list(enumerables, {:cont, acc}, fun, zip_fun) do
+    case zip_list_heads_tails(enumerables, [], []) do
+      {heads, tails} -> zip_list(tails, fun.(zip_fun.(heads), acc), fun, zip_fun)
+      :error -> {:done, acc}
+    end
+  end
+
+  defp zip_list_heads_tails([[head | tail] | rest], heads, tails) do
+    zip_list_heads_tails(rest, [head | heads], [tail | tails])
+  end
+
+  defp zip_list_heads_tails([[] | _rest], _heads, _tails) do
+    :error
+  end
+
+  defp zip_list_heads_tails([], heads, tails) do
+    {:lists.reverse(heads), :lists.reverse(tails)}
+  end
+
+  defp zip_enum(enumerables, acc, fun, zip_fun) do
+    step = fn x, acc ->
+      {:suspend, :lists.reverse([x | acc])}
+    end
 
     enum_funs =
       Enum.map(enumerables, fn enum ->
         {&Enumerable.reduce(enum, &1, step), [], :cont}
       end)
 
-    do_zip(enum_funs, acc, fun)
+    do_zip_enum(enum_funs, acc, fun, zip_fun)
   end
 
-  # This implementation of do_zip/3 works for any number of
-  # streams to zip, even if right now zip/2 only zips two streams.
-
-  defp do_zip(zips, {:halt, acc}, _fun) do
+  # This implementation of do_zip_enum/4 works for any number of streams to zip
+  defp do_zip_enum(zips, {:halt, acc}, _fun, _zip_fun) do
     do_zip_close(zips)
     {:halted, acc}
   end
 
-  defp do_zip(zips, {:suspend, acc}, fun) do
-    {:suspended, acc, &do_zip(zips, &1, fun)}
+  defp do_zip_enum(zips, {:suspend, acc}, fun, zip_fun) do
+    {:suspended, acc, &do_zip_enum(zips, &1, fun, zip_fun)}
   end
 
-  defp do_zip([], {:cont, acc}, _callback) do
+  defp do_zip_enum([], {:cont, acc}, _callback, _zip_fun) do
     {:done, acc}
   end
 
-  defp do_zip(zips, {:cont, acc}, callback) do
+  defp do_zip_enum(zips, {:cont, acc}, callback, zip_fun) do
     try do
-      do_zip_next_tuple(zips, acc, callback, [], [])
+      do_zip_next(zips, acc, callback, [], [], zip_fun)
     catch
       kind, reason ->
         do_zip_close(zips)
         :erlang.raise(kind, reason, __STACKTRACE__)
     else
       {:next, buffer, acc} ->
-        do_zip(buffer, acc, callback)
+        do_zip_enum(buffer, acc, callback, zip_fun)
 
       {:done, _acc} = other ->
         other
     end
   end
 
-  # do_zip_next_tuple/5 computes the next tuple formed by
+  # do_zip_next/6 computes the next tuple formed by
   # the next element of each zipped stream.
-
-  defp do_zip_next_tuple([{_, [], :halt} | zips], acc, _callback, _yielded_elems, buffer) do
+  defp do_zip_next(
+         [{_, [], :halt} | zips],
+         acc,
+         _callback,
+         _yielded_elems,
+         buffer,
+         _zip_fun
+       ) do
     do_zip_close(:lists.reverse(buffer, zips))
     {:done, acc}
   end
 
-  defp do_zip_next_tuple([{fun, [], :cont} | zips], acc, callback, yielded_elems, buffer) do
+  defp do_zip_next([{fun, [], :cont} | zips], acc, callback, yielded_elems, buffer, zip_fun) do
     case fun.({:cont, []}) do
       {:suspended, [elem | next_acc], fun} ->
         next_buffer = [{fun, next_acc, :cont} | buffer]
-        do_zip_next_tuple(zips, acc, callback, [elem | yielded_elems], next_buffer)
+        do_zip_next(zips, acc, callback, [elem | yielded_elems], next_buffer, zip_fun)
 
       {_, [elem | next_acc]} ->
         next_buffer = [{fun, next_acc, :halt} | buffer]
-        do_zip_next_tuple(zips, acc, callback, [elem | yielded_elems], next_buffer)
+        do_zip_next(zips, acc, callback, [elem | yielded_elems], next_buffer, zip_fun)
 
       {_, []} ->
         # The current zipped stream terminated, so we close all the streams
@@ -1209,28 +1411,30 @@ defmodule Stream do
     end
   end
 
-  defp do_zip_next_tuple([{fun, zip_acc, zip_op} | zips], acc, callback, yielded_elems, buffer) do
+  defp do_zip_next(
+         [{fun, zip_acc, zip_op} | zips],
+         acc,
+         callback,
+         yielded_elems,
+         buffer,
+         zip_fun
+       ) do
     [elem | rest] = zip_acc
     next_buffer = [{fun, rest, zip_op} | buffer]
-    do_zip_next_tuple(zips, acc, callback, [elem | yielded_elems], next_buffer)
+    do_zip_next(zips, acc, callback, [elem | yielded_elems], next_buffer, zip_fun)
   end
 
-  defp do_zip_next_tuple([] = _zips, acc, callback, yielded_elems, buffer) do
+  defp do_zip_next([] = _zips, acc, callback, yielded_elems, buffer, zip_fun) do
     # "yielded_elems" is a reversed list of results for the current iteration of
-    # zipping: it needs to be reversed and converted to a tuple to have the next
-    # tuple in the list resulting from zipping.
-    zipped = List.to_tuple(:lists.reverse(yielded_elems))
-    {:next, :lists.reverse(buffer), callback.(zipped, acc)}
+    # zipping. That is to say, the nth element from each of the enums being zipped.
+    # It needs to be reversed and passed to the zipping function so it can do it's thing.
+    {:next, :lists.reverse(buffer), callback.(zip_fun.(:lists.reverse(yielded_elems)), acc)}
   end
 
   defp do_zip_close(zips) do
     :lists.foreach(fn {fun, _, _} -> fun.({:halt, []}) end, zips)
   end
 
-  defp do_zip_step(x, acc) do
-    {:suspend, :lists.reverse([x | acc])}
-  end
-
   ## Sources
 
   @doc """
@@ -1248,7 +1452,7 @@ defmodule Stream do
   def cycle(enumerable)
 
   def cycle([]) do
-    raise ArgumentError, "cannot cycle over empty enumerable"
+    raise ArgumentError, "cannot cycle over an empty enumerable"
   end
 
   def cycle(enumerable) when is_list(enumerable) do
@@ -1297,7 +1501,7 @@ defmodule Stream do
     fn acc ->
       case reduce.(acc) do
         {state, []} when state in [:done, :halted] ->
-          raise ArgumentError, "cannot cycle over empty enumerable"
+          raise ArgumentError, "cannot cycle over an empty enumerable"
 
         other ->
           other
@@ -1333,9 +1537,9 @@ defmodule Stream do
   ## Examples
 
       # Although not necessary, let's seed the random algorithm
-      iex> :rand.seed(:exrop, {1, 2, 3})
+      iex> :rand.seed(:exsss, {1, 2, 3})
       iex> Stream.repeatedly(&:rand.uniform/0) |> Enum.take(3)
-      [0.7498295129076106, 0.06161655489244533, 0.7924073127680873]
+      [0.5455598952593053, 0.6039309974353404, 0.6684893034823949]
 
   """
   @spec repeatedly((() -> element)) :: Enumerable.t()
@@ -1584,7 +1788,7 @@ defmodule Stream do
 end
 
 defimpl Enumerable, for: Stream do
-  @compile :inline_list_funs
+  @compile :inline_list_funcs
 
   def count(_lazy), do: {:error, __MODULE__}
 
@@ -1621,7 +1825,7 @@ defimpl Enumerable, for: Stream do
   defp do_done({reason, [acc | _]}, nil), do: {reason, acc}
 
   defp do_done({reason, [acc | t]}, {done, fun}) do
-    [h | _] = Enum.reverse(t)
+    [h | _] = :lists.reverse(t)
 
     case done.([acc, h], fun) do
       {:cont, [acc | _]} -> {reason, acc}
@@ -1635,7 +1839,7 @@ defimpl Inspect, for: Stream do
   import Inspect.Algebra
 
   def inspect(%{enum: enum, funs: funs}, opts) do
-    inner = [enum: enum, funs: Enum.reverse(funs)]
+    inner = [enum: enum, funs: :lists.reverse(funs)]
     concat(["#Stream<", to_doc(inner, opts), ">"])
   end
 end
