@@ -50,13 +50,13 @@ defmodule SystemTest do
 
   @test_var "SYSTEM_ELIXIR_ENV_TEST_VAR"
 
-  test "*_env/*" do
+  test "get_env/put_env/delete_env" do
     assert System.get_env(@test_var) == nil
     assert System.get_env(@test_var, "SAMPLE") == "SAMPLE"
     assert System.fetch_env(@test_var) == :error
 
     message = "could not fetch environment variable #{inspect(@test_var)} because it is not set"
-    assert_raise ArgumentError, message, fn -> System.fetch_env!(@test_var) end
+    assert_raise System.EnvError, message, fn -> System.fetch_env!(@test_var) end
 
     System.put_env(@test_var, "SAMPLE")
 
@@ -68,14 +68,22 @@ defmodule SystemTest do
     System.delete_env(@test_var)
     assert System.get_env(@test_var) == nil
 
-    System.put_env(%{@test_var => "OTHER_SAMPLE"})
-    assert System.get_env(@test_var) == "OTHER_SAMPLE"
-
     assert_raise ArgumentError, ~r[cannot execute System.put_env/2 for key with \"=\"], fn ->
       System.put_env("FOO=BAR", "BAZ")
     end
   end
 
+  test "put_env/2" do
+    System.put_env(%{@test_var => "MAP_STRING"})
+    assert System.get_env(@test_var) == "MAP_STRING"
+
+    System.put_env([{String.to_atom(@test_var), "KW_ATOM"}])
+    assert System.get_env(@test_var) == "KW_ATOM"
+
+    System.put_env([{String.to_atom(@test_var), nil}])
+    assert System.get_env(@test_var) == nil
+  end
+
   test "cmd/2 raises for null bytes" do
     assert_raise ArgumentError, ~r"cannot execute System.cmd/3 for program with null byte", fn ->
       System.cmd("null\0byte", [])
@@ -96,22 +104,21 @@ defmodule SystemTest do
     end
 
     test "cmd/3 (with options)" do
-      assert {["hello\r\n"], 0} =
-               System.cmd(
-                 "cmd",
-                 ~w[/c echo hello],
-                 into: [],
-                 cd: File.cwd!(),
-                 env: %{"foo" => "bar", "baz" => nil},
-                 arg0: "echo",
-                 stderr_to_stdout: true,
-                 parallelism: true
-               )
+      opts = [
+        into: [],
+        cd: File.cwd!(),
+        env: %{"foo" => "bar", "baz" => nil},
+        arg0: "echo",
+        stderr_to_stdout: true,
+        parallelism: true
+      ]
+
+      assert {["hello\r\n"], 0} = System.cmd("cmd", ~w[/c echo hello], opts)
     end
 
     @echo "echo-elixir-test"
     @tag :tmp_dir
-    test "cmd/2 with absolute and relative paths", config do
+    test "cmd/3 with absolute and relative paths", config do
       echo = Path.join(config.tmp_dir, @echo)
       File.mkdir_p!(Path.dirname(echo))
       File.cp!(System.find_executable("cmd"), echo)
@@ -128,6 +135,22 @@ defmodule SystemTest do
                  System.cmd(Path.join(File.cwd!(), @echo), ~w[/c echo hello], [{:arg0, "echo"}])
       end)
     end
+
+    test "shell/1" do
+      assert {"hello\r\n", 0} = System.shell("echo hello")
+    end
+
+    test "shell/2 (with options)" do
+      opts = [
+        into: [],
+        cd: File.cwd!(),
+        env: %{"foo" => "bar", "baz" => nil},
+        stderr_to_stdout: true,
+        parallelism: true
+      ]
+
+      assert {["bar\r\n"], 0} = System.shell("echo %foo%", opts)
+    end
   end
 
   describe "Unix" do
@@ -152,7 +175,7 @@ defmodule SystemTest do
 
     @echo "echo-elixir-test"
     @tag :tmp_dir
-    test "cmd/2 with absolute and relative paths", config do
+    test "cmd/3 with absolute and relative paths", config do
       echo = Path.join(config.tmp_dir, @echo)
       File.mkdir_p!(Path.dirname(echo))
       File.cp!(System.find_executable("echo"), echo)
@@ -169,6 +192,67 @@ defmodule SystemTest do
                  System.cmd(Path.join(File.cwd!(), @echo), ["hello"], [{:arg0, "echo"}])
       end)
     end
+
+    test "shell/1" do
+      assert {"hello\n", 0} = System.shell("echo hello")
+    end
+
+    test "shell/1 with interpolation" do
+      assert {"1\n2\n", 0} = System.shell("x=1; echo $x; echo '2'")
+    end
+
+    @tag timeout: 1_000
+    test "shell/1 returns when command awaits input" do
+      assert {"", 0} = System.shell("cat", close_stdin: true)
+    end
+
+    test "shell/1 with comment" do
+      assert {"1\n", 0} = System.shell("echo '1' # comment")
+    end
+
+    test "shell/2 (with options)" do
+      opts = [
+        into: [],
+        cd: File.cwd!(),
+        env: %{"foo" => "bar", "baz" => nil},
+        stderr_to_stdout: true
+      ]
+
+      assert {["bar\n"], 0} = System.shell("echo $foo", opts)
+    end
+  end
+
+  @tag :unix
+  test "vm signals" do
+    assert System.trap_signal(:sigquit, :example, fn -> :ok end) == {:ok, :example}
+    assert System.trap_signal(:sigquit, :example, fn -> :ok end) == {:error, :already_registered}
+    assert {:ok, ref} = System.trap_signal(:sigquit, fn -> :ok end)
+
+    assert System.untrap_signal(:sigquit, :example) == :ok
+    assert System.trap_signal(:sigquit, :example, fn -> :ok end) == {:ok, :example}
+    assert System.trap_signal(:sigquit, ref, fn -> :ok end) == {:error, :already_registered}
+
+    assert System.untrap_signal(:sigusr1, :example) == {:error, :not_found}
+    assert System.untrap_signal(:sigquit, :example) == :ok
+    assert System.untrap_signal(:sigquit, :example) == {:error, :not_found}
+    assert System.untrap_signal(:sigquit, ref) == :ok
+    assert System.untrap_signal(:sigquit, ref) == {:error, :not_found}
+  end
+
+  @tag :unix
+  test "os signals" do
+    parent = self()
+
+    assert System.trap_signal(:sighup, :example, fn ->
+             send(parent, :sighup_called)
+             :ok
+           end) == {:ok, :example}
+
+    {"", 0} = System.cmd("kill", ["-s", "hup", System.pid()])
+
+    assert_receive :sighup_called
+  after
+    System.untrap_signal(:sighup, :example)
   end
 
   test "find_executable/1" do
