@@ -40,12 +40,13 @@ defmodule IEx.Helpers do
     * `pid/3`          - creates a PID with the 3 integer arguments passed
     * `port/1`         - creates a port from a string
     * `port/2`         - creates a port with the 2 non-negative integers passed
-    * `ref/1`          - creates a reference from a string
-    * `ref/4`          - creates a reference with the 4 integer arguments passed
     * `pwd/0`          - prints the current working directory
     * `r/1`            - recompiles the given module's source file
     * `recompile/0`    - recompiles the current project
+    * `ref/1`          - creates a reference from a string
+    * `ref/4`          - creates a reference with the 4 integer arguments passed
     * `runtime_info/0` - prints runtime info (versions, memory usage, stats)
+    * `t/1`            - prints the types for the given module or function
     * `v/0`            - retrieves the last value from the history
     * `v/1`            - retrieves the nth value from the history
 
@@ -68,14 +69,17 @@ defmodule IEx.Helpers do
   import IEx, only: [dont_display_result: 0]
 
   @doc """
-  Recompiles the current Mix application.
+  Recompiles the current Mix project.
 
   This helper only works when IEx is started with a Mix
-  project, for example, `iex -S mix`. The application is
-  not restarted after compilation, which means any long
-  running process may crash as any changed module will be
-  temporarily removed and recompiled, without going through
-  the proper code changes callback.
+  project, for example, `iex -S mix`. Note this function
+  simply recompiles Elixir modules, without reloading
+  configuration, recompiling dependencies, or restarting
+  applications.
+
+  Therefore, any long running process may crash on recompilation,
+  as changed modules will be temporarily removed and recompiled,
+  without going through the proper code change callback.
 
   If you want to reload a single module, consider using
   `r(ModuleName)` instead.
@@ -103,7 +107,7 @@ defmodule IEx.Helpers do
 
       {result, _} = Mix.Task.run("compile", arguments)
 
-      # Reenable consolidation and allow them to be loaded.
+      # Re-enable consolidation and allow them to be loaded.
       Code.prepend_path(consolidation)
       purge_protocols(consolidation)
 
@@ -252,24 +256,32 @@ defmodule IEx.Helpers do
   and falls back to `EDITOR` if the former is not available.
 
   By default, it attempts to open the file and line using the
-  `file:line` notation. For example, if your editor is called
-  `subl`, it will open the file as:
+  `file:line` notation. For example, for Sublime Text you can
+  set it as:
+
+      ELIXIR_EDITOR="subl"
+
+  Which will then try to open it as:
 
       subl path/to/file:line
 
+  For Visual Studio Code, once enabled on the command line,
+  you can set it to:
+
+      ELIXIR_EDITOR="code --goto"
+
   It is important that you choose an editor command that does
   not block nor that attempts to run an editor directly in the
   terminal. Command-line based editors likely need extra
   configuration so they open up the given file and line in a
   separate window.
 
-  Custom editors are supported by using the `__FILE__` and
-  `__LINE__` notations, for example:
+  For more complex use cases, you can use the `__FILE__` and
+  `__LINE__` notations to explicitly interpolate the file and
+  line into the command:
 
       ELIXIR_EDITOR="my_editor +__LINE__ __FILE__"
 
-  and Elixir will properly interpolate values.
-
   Since this function prints the result returned by the editor,
   `ELIXIR_EDITOR` can be set "echo" if you prefer to display the
   location rather than opening it.
@@ -391,10 +403,15 @@ defmodule IEx.Helpers do
   end
 
   @doc """
-  Recompiles and reloads the given `module`.
+  Recompiles and reloads the given `module` or `modules`.
 
-  Please note that all the modules defined in the same
-  file as `module` are recompiled and reloaded.
+  Please note that all the modules defined in the same file as
+  `modules` are recompiled and reloaded. If you want to reload
+  multiple modules, it is best to reload them at the same time,
+  such as in `r [Foo, Bar]`. This is important to avoid false
+  warnings, since the module is only reloaded in memory and its
+  latest information is not persisted to disk. See the "In-memory
+  reloading" section below.
 
   This function is meant to be used for development and
   debugging purposes. Do not depend on it in production code.
@@ -406,35 +423,49 @@ defmodule IEx.Helpers do
   in disk, probably the one where the first definition of the module
   came from, does not change at all.
 
-  Since typespecs and docs are loaded from the .beam file (they
-  are not loaded in memory with the module because there is no need
-  for them to be in memory), they are not reloaded when you reload
-  the module.
+  Since docs, typespecs, and exports information are loaded from the
+  .beam file, they are not reloaded when you invoke this function.
   """
-  def r(module) when is_atom(module) do
-    {:reloaded, module, do_r(module)}
-  end
+  def r(module_or_modules) do
+    modules = List.wrap(module_or_modules)
 
-  defp do_r(module) do
-    unless Code.ensure_loaded?(module) do
-      raise ArgumentError, "could not load nor find module: #{inspect(module)}"
-    end
+    sources =
+      Enum.map(modules, fn module ->
+        unless Code.ensure_loaded?(module) do
+          raise ArgumentError, "could not load nor find module: #{inspect(module)}"
+        end
 
-    source = source(module)
+        source = source(module)
 
-    cond do
-      source == nil ->
-        raise ArgumentError, "could not find source for module: #{inspect(module)}"
+        cond do
+          source == nil ->
+            raise ArgumentError, "could not find source for module: #{inspect(module)}"
 
-      not File.exists?(source) ->
-        raise ArgumentError, "could not find source (#{source}) for module: #{inspect(module)}"
+          not File.exists?(source) ->
+            raise ArgumentError,
+                  "could not find source (#{source}) for module: #{inspect(module)}"
 
-      String.ends_with?(source, ".erl") ->
-        [compile_erlang(source) |> elem(0)]
+          true ->
+            source
+        end
+      end)
 
-      true ->
-        Enum.map(Code.compile_file(source), fn {name, _} -> name end)
-    end
+    {erlang, elixir} = Enum.split_with(sources, &String.ends_with?(&1, ".erl"))
+
+    erlang =
+      for source <- erlang do
+        compile_erlang(source) |> elem(0)
+      end
+
+    elixir =
+      if elixir != [] do
+        {:ok, modules, _warning} = Kernel.ParallelCompiler.compile(elixir)
+        modules
+      else
+        []
+      end
+
+    {:reloaded, erlang ++ elixir}
   end
 
   @doc """
@@ -492,21 +523,29 @@ defmodule IEx.Helpers do
     :code.get_path()
     |> Protocol.extract_protocols()
     |> Enum.uniq()
-    |> Enum.reject(fn protocol -> is_nil(protocol.impl_for(term)) end)
+    |> Enum.filter(fn protocol ->
+      Code.ensure_loaded?(protocol) and function_exported?(protocol, :impl_for, 1) and
+        protocol.impl_for(term) != nil
+    end)
     |> Enum.sort()
     |> Enum.map_join(", ", &inspect/1)
   end
 
-  @runtime_info_topics [:system, :memory, :allocators, :limits, :applications]
+  @runtime_info_topics [:system, :memory, :limits, :applications]
   @doc """
   Prints VM/runtime information such as versions, memory usage and statistics.
+
   Additional topics are available via `runtime_info/1`.
+
+  For more metrics, info, and debugging facilities, see the
+  [Recon](https://github.com/ferd/recon) project.
   """
   @doc since: "1.5.0"
   def runtime_info(), do: runtime_info([:system, :memory, :limits])
 
   @doc """
   Just like `runtime_info/0`, except accepts topic or a list of topics.
+
   For example, topic `:applications` will list the applications loaded.
   """
   @doc since: "1.5.0"
@@ -543,7 +582,7 @@ defmodule IEx.Helpers do
     print_pane("System and architecture")
 
     print_entry("Elixir version", System.version())
-    print_entry("Erlang/OTP version", :erlang.system_info(:otp_release))
+    print_entry("Erlang/OTP version", System.otp_release())
     print_entry("ERTS version", :erlang.system_info(:version))
     print_entry("Compiled for", :erlang.system_info(:system_architecture))
     print_entry("Schedulers", :erlang.system_info(:schedulers))
@@ -560,31 +599,6 @@ defmodule IEx.Helpers do
     print_memory("Processes", :processes)
   end
 
-  defp print_runtime_info_topic(:allocators) do
-    print_pane("Allocators")
-
-    areas = :erlang.system_info(:allocated_areas)
-
-    print_allocator(areas, "Atom Space", :atom_space)
-    print_allocator(areas, "Atom Table", :atom_table)
-    print_allocator(areas, "BIF Timer", :bif_timer)
-    print_allocator(areas, "Bits Bufs Size", :bits_bufs_size)
-    print_allocator(areas, "Dist Table", :dist_table)
-    print_allocator(areas, "ETS Misc", :ets_misc)
-    print_allocator(areas, "Export List", :export_list)
-    print_allocator(areas, "Export Table", :export_table)
-    print_allocator(areas, "Function Table", :fun_table)
-    print_allocator(areas, "Loaded Code", :loaded_code)
-    print_allocator(areas, "Module Refs", :module_refs)
-    print_allocator(areas, "Module Table", :module_table)
-    print_allocator(areas, "Node Table", :node_table)
-    print_allocator(areas, "Port Table", :port_table)
-    print_allocator(areas, "Process Table", :process_table)
-    print_allocator(areas, "Register Table", :register_table)
-    print_allocator(areas, "Static", :static)
-    print_allocator(areas, "System Misc", :sys_misc)
-  end
-
   defp print_runtime_info_topic(:limits) do
     print_pane("Statistics / limits")
     print_uptime()
@@ -634,7 +648,7 @@ defmodule IEx.Helpers do
     IO.puts("#{pad_key(key)}#{min} / #{max} (#{percentage}% used)")
   end
 
-  defp get_stat(:ets_count), do: length(:ets.all())
+  defp get_stat(:ets_count), do: :erlang.system_info(:ets_count)
   defp get_stat(other), do: :erlang.system_info(other)
 
   defp print_memory(key, memory) do
@@ -642,19 +656,6 @@ defmodule IEx.Helpers do
     IO.puts("#{pad_key(key)}#{format_bytes(value)}")
   end
 
-  defp print_allocator(allocated_areas, key, probe) do
-    case List.keyfind(allocated_areas, probe, 0) do
-      {_, allocated, used} ->
-        IO.puts("#{pad_key(key)}#{format_bytes(allocated)} (#{format_bytes(used)} used)")
-
-      {_, allocated} ->
-        IO.puts("#{pad_key(key)}#{format_bytes(allocated)}")
-
-      _ ->
-        IO.puts("#{pad_key(key)}N/A")
-    end
-  end
-
   defp format_bytes(bytes) when is_integer(bytes) do
     cond do
       bytes >= memory_unit(:GB) -> format_bytes(bytes, :GB)
@@ -790,7 +791,7 @@ defmodule IEx.Helpers do
     # print items in multiple columns (2 columns in the worst case)
     lengths = Enum.map(list, &String.length(&1))
     max_length = max_length(lengths)
-    offset = min(max_length, 30) + 5
+    offset = min(max_length, 30) + 4
     print_table(list, printer, offset)
   end
 
@@ -804,8 +805,11 @@ defmodule IEx.Helpers do
           length
         end
 
-      IO.write(printer.(item, offset))
-      length + offset
+      printed = printer.(item, offset)
+      actual_offset = String.length(printed) + 1
+
+      IO.write(printed <> " ")
+      length + actual_offset
     end)
 
     IO.puts("")
@@ -847,6 +851,9 @@ defmodule IEx.Helpers do
   until the next breakpoint, which will automatically yield control
   back to IEx without requesting permission to pry.
 
+  If you simply want to move to the next line of the current breakpoint,
+  use `n/0` or `next/0` instead.
+
   If the running process terminates, a new IEx session is
   started.
 
@@ -857,14 +864,51 @@ defmodule IEx.Helpers do
   @doc since: "1.5.0"
   def continue do
     if iex_server = Process.get(:iex_server) do
-      send(iex_server, {:continue, self()})
+      send(iex_server, {:continue, self(), false})
     end
 
     dont_display_result()
   end
 
   @doc """
-  Macro-based shortcut for `IEx.break!/4`.
+  Goes to the next line of the current breakpoint.
+
+  This is usually called by sessions started with `IEx.break!/4`.
+  If instead of the next line you want to move to the next breakpoint,
+  call `continue/0` instead.
+
+  While the process executes, the user will no longer have
+  control of the shell. If you would rather start a new shell,
+  use `respawn/0` instead.
+  """
+  @doc since: "1.14.0"
+  def next do
+    if iex_server = Process.get(:iex_server) do
+      send(iex_server, {:continue, self(), true})
+    end
+
+    dont_display_result()
+  end
+
+  @doc """
+  A shortcut for `next/0`.
+  """
+  @doc since: "1.14.0"
+  def n do
+    next()
+  end
+
+  @doc """
+  Sets up a breakpoint in the AST of shape `Module.function/arity`
+  with the given number of `stops`.
+
+  See `IEx.break!/4` for a complete description of breakpoints
+  in IEx.
+
+  ## Examples
+
+      break! URI.decode_query/2
+
   """
   @doc since: "1.5.0"
   defmacro break!(ast, stops \\ 1) do
@@ -880,6 +924,11 @@ defmodule IEx.Helpers do
 
   See `IEx.break!/4` for a complete description of breakpoints
   in IEx.
+
+  ## Examples
+
+      break! URI, :decode_query, 2
+
   """
   @doc since: "1.5.0"
   defdelegate break!(module, function, arity, stops \\ 1), to: IEx
@@ -1083,11 +1132,10 @@ defmodule IEx.Helpers do
   end
 
   @doc """
-  Injects the contents of the file at `path` as if it was typed into
-  the shell.
+  Injects the contents of the file at `path`.
 
   This would be the equivalent of getting all of the file contents and
-  packing it all into a single line in IEx and executing it.
+  pasting it all at once in IEx and executing it.
 
   By default, the contents of a `.iex.exs` file in the same directory
   as you are starting IEx are automatically imported. See the section
@@ -1189,18 +1237,28 @@ defmodule IEx.Helpers do
   defp history, do: Process.get(:iex_history)
 
   @doc """
-  Creates a PID from `string`.
+  Creates a PID from `string` or `atom`.
 
   ## Examples
 
       iex> pid("0.21.32")
       #PID<0.21.32>
 
+      iex> pid(:init)
+      #PID<0.0.0>
+
   """
   def pid(string) when is_binary(string) do
     :erlang.list_to_pid('<#{string}>')
   end
 
+  def pid(name) when is_atom(name) do
+    case Process.whereis(name) do
+      p when is_pid(p) -> p
+      _ -> raise ArgumentError, "could not find registered process with name: #{inspect(name)}"
+    end
+  end
+
   @doc """
   Creates a PID with 3 non-negative integers passed as arguments
   to the function.
@@ -1316,8 +1374,8 @@ defmodule IEx.Helpers do
 
   """
   def nl(nodes \\ Node.list(), module) when is_list(nodes) and is_atom(module) do
-    case :code.get_object_code(module) do
-      {^module, bin, beam_path} ->
+    case get_beam_and_path(module) do
+      {bin, beam_path} ->
         results =
           for node <- nodes do
             case :rpc.call(node, :code, :load_binary, [module, beam_path, bin]) do
@@ -1334,4 +1392,13 @@ defmodule IEx.Helpers do
         {:error, :nofile}
     end
   end
+
+  defp get_beam_and_path(module) do
+    with {^module, beam, filename} <- :code.get_object_code(module),
+         {:ok, ^module} <- beam |> :beam_lib.info() |> Keyword.fetch(:module) do
+      {beam, filename}
+    else
+      _ -> :error
+    end
+  end
 end
