@@ -2,8 +2,22 @@ defmodule IEx.State do
   @moduledoc false
   # This state is exchanged between IEx.Server and
   # IEx.Evaluator which is why it is a struct.
-  defstruct cache: '', counter: 1, prefix: "iex", on_eof: :stop_evaluator, evaluator_options: []
-  @type t :: %__MODULE__{}
+  defstruct parser_state: "",
+            counter: 1,
+            prefix: "iex",
+            on_eof: :stop_evaluator,
+            evaluator_options: [],
+            previous_state: nil
+
+  @type t :: %{
+          __struct__: __MODULE__,
+          parser_state: binary(),
+          counter: pos_integer(),
+          prefix: binary(),
+          on_eof: :stop_evaluator | :halt,
+          evaluator_options: keyword(),
+          previous_state: nil | t()
+        }
 end
 
 defmodule IEx.Server do
@@ -58,6 +72,7 @@ defmodule IEx.Server do
     receive do
       {:take_over, take_pid, take_ref, take_location, take_whereami, take_opts} ->
         if take_over?(take_pid, take_ref, take_location, take_whereami, take_opts) do
+          take_opts = Keyword.put(take_opts, :previous_state, iex_state(opts))
           run_without_registration(take_opts)
         else
           shell_loop(opts, pid, ref)
@@ -82,7 +97,7 @@ defmodule IEx.Server do
     )
 
     evaluator = start_evaluator(opts)
-    loop(iex_state(opts), evaluator, Process.monitor(evaluator))
+    loop(iex_state(opts), :ok, evaluator, Process.monitor(evaluator))
   end
 
   # Starts an evaluator using the provided options.
@@ -103,7 +118,7 @@ defmodule IEx.Server do
   defp stop_evaluator(evaluator, evaluator_ref) do
     Process.delete(:evaluator)
     Process.demonitor(evaluator_ref, [:flush])
-    send(evaluator, {:done, self()})
+    send(evaluator, {:done, self(), false})
     :ok
   end
 
@@ -114,12 +129,12 @@ defmodule IEx.Server do
     run_without_registration(opts)
   end
 
-  defp loop(state, evaluator, evaluator_ref) do
+  defp loop(state, prompt, evaluator, evaluator_ref) do
     self_pid = self()
     counter = state.counter
 
     {prompt_type, prefix} =
-      if state.cache != [] do
+      if prompt == :incomplete do
         {:continuation_prompt, "..."}
       else
         {:prompt, state.prefix}
@@ -144,7 +159,7 @@ defmodule IEx.Server do
       # Triggered by pressing "i" as the job control switch
       {:input, ^input, {:error, :interrupted}} ->
         io_error("** (EXIT) interrupted")
-        loop(%{state | cache: ''}, evaluator, evaluator_ref)
+        loop(%{state | parser_state: ""}, :new, evaluator, evaluator_ref)
 
       # Triggered when IO dies while waiting for input
       {:input, ^input, {:error, :terminated}} ->
@@ -159,8 +174,8 @@ defmodule IEx.Server do
 
   defp wait_eval(state, evaluator, evaluator_ref) do
     receive do
-      {:evaled, ^evaluator, new_state} ->
-        loop(new_state, evaluator, evaluator_ref)
+      {:evaled, ^evaluator, status, new_state} ->
+        loop(new_state, status, evaluator, evaluator_ref)
 
       msg ->
         handle_take_over(msg, state, evaluator, evaluator_ref, nil, fn state ->
@@ -196,12 +211,14 @@ defmodule IEx.Server do
 
         if take_over?(take_pid, take_ref, true) do
           kill_input(input)
-          loop(iex_state(take_opts), evaluator, evaluator_ref)
+          take_opts = Keyword.put(take_opts, :previous_state, state)
+          loop(iex_state(take_opts), :ok, evaluator, evaluator_ref)
         else
           callback.(state)
         end
 
       take_over?(take_pid, take_ref, take_location, take_whereami, take_opts) ->
+        take_opts = Keyword.put(take_opts, :previous_state, state)
         rerun(take_opts, evaluator, evaluator_ref, input)
 
       true ->
@@ -224,18 +241,19 @@ defmodule IEx.Server do
     Process.exit(evaluator, :kill)
     Process.demonitor(evaluator_ref, [:flush])
     evaluator = start_evaluator(state.evaluator_options)
-    loop(%{state | cache: ''}, evaluator, Process.monitor(evaluator))
+    loop(%{state | parser_state: ""}, :ok, evaluator, Process.monitor(evaluator))
   end
 
   defp handle_take_over(
          {:EXIT, pid, reason},
          state,
-         _evaluator,
-         _evaluator_ref,
+         evaluator,
+         evaluator_ref,
          _input,
          callback
        ) do
     if pid == Process.group_leader() do
+      stop_evaluator(evaluator, evaluator_ref)
       exit(reason)
     else
       callback.(state)
@@ -243,12 +261,24 @@ defmodule IEx.Server do
   end
 
   defp handle_take_over({:respawn, evaluator}, state, evaluator, evaluator_ref, input, _callback) do
-    rerun(state.evaluator_options, evaluator, evaluator_ref, input)
+    opts =
+      if state.previous_state,
+        do: state.previous_state.evaluator_options,
+        else: state.evaluator_options
+
+    rerun(opts, evaluator, evaluator_ref, input)
   end
 
-  defp handle_take_over({:continue, evaluator}, state, evaluator, evaluator_ref, input, _callback) do
+  defp handle_take_over(
+         {:continue, evaluator, next?},
+         state,
+         evaluator,
+         evaluator_ref,
+         input,
+         _callback
+       ) do
     kill_input(input)
-    send(evaluator, {:done, self()})
+    send(evaluator, {:done, self(), next?})
     wait_take_over(state, evaluator, evaluator_ref)
   end
 
@@ -260,7 +290,12 @@ defmodule IEx.Server do
          input,
          _callback
        ) do
-    rerun(state.evaluator_options, evaluator, evaluator_ref, input)
+    opts =
+      if state.previous_state,
+        do: state.previous_state.evaluator_options,
+        else: state.evaluator_options
+
+    rerun(opts, evaluator, evaluator_ref, input)
   end
 
   defp handle_take_over(
@@ -281,7 +316,12 @@ defmodule IEx.Server do
         io_error("** (IEx.Error) #{type} when printing EXIT message: #{inspect(detail)}")
     end
 
-    rerun(state.evaluator_options, evaluator, evaluator_ref, input)
+    opts =
+      if state.previous_state,
+        do: state.previous_state.evaluator_options,
+        else: state.evaluator_options
+
+    rerun(opts, evaluator, evaluator_ref, input)
   end
 
   defp handle_take_over(_, state, _evaluator, _evaluator_ref, _input, callback) do
@@ -325,6 +365,7 @@ defmodule IEx.Server do
     %IEx.State{
       prefix: prefix,
       on_eof: on_eof,
+      previous_state: Keyword.get(opts, :previous_state),
       evaluator_options: Keyword.take(opts, [:dot_iex_path])
     }
   end
@@ -333,9 +374,29 @@ defmodule IEx.Server do
 
   defp io_get(pid, prompt_type, prefix, counter) do
     prompt = prompt(prompt_type, prefix, counter)
-    send(pid, {:input, self(), IO.gets(:stdio, prompt)})
+    send(pid, {:input, self(), io_get(prompt)})
   end
 
+  defp io_get(prompt) do
+    pid = Process.group_leader()
+    ref = Process.monitor(pid)
+    command = {:get_until, :unicode, prompt, __MODULE__, :__io_get__, []}
+    send(pid, {:io_request, self(), ref, command})
+
+    receive do
+      {:io_reply, ^ref, reply} ->
+        Process.demonitor(ref, [:flush])
+        reply
+
+      {:DOWN, ^ref, _, _, _} ->
+        {:error, :terminated}
+    end
+  end
+
+  @doc false
+  def __io_get__([], :eof), do: {:done, :eof, []}
+  def __io_get__([], chars), do: {:done, List.to_string(chars), []}
+
   defp prompt(prompt_type, prefix, counter) do
     {mode, prefix} =
       if Node.alive?() do
