@@ -9,23 +9,17 @@ defmodule IEx.AutocompleteTest do
     :ok
   end
 
-  defmodule MyServer do
-    def evaluator do
-      {Process.get(:evaluator), self()}
-    end
-  end
-
   defp eval(line) do
     ExUnit.CaptureIO.capture_io(fn ->
-      {evaluator, _} = MyServer.evaluator()
+      evaluator = Process.get(:evaluator)
       Process.group_leader(evaluator, Process.group_leader())
       send(evaluator, {:eval, self(), line <> "\n", %IEx.State{}})
-      assert_receive {:evaled, _, _}
+      assert_receive {:evaled, _, _, _}
     end)
   end
 
   defp expand(expr) do
-    IEx.Autocomplete.expand(Enum.reverse(expr), MyServer)
+    IEx.Autocomplete.expand(Enum.reverse(expr), self())
   end
 
   test "Erlang module completion" do
@@ -34,19 +28,19 @@ defmodule IEx.AutocompleteTest do
 
   test "Erlang module no completion" do
     assert expand(':unknown') == {:no, '', []}
-    assert expand('Enum:') == {:no, '', []}
   end
 
   test "Erlang module multiple values completion" do
-    {:yes, '', list} = expand(':user')
-    assert 'user' in list
-    assert 'user_drv' in list
+    {:yes, '', list} = expand(':logger')
+    assert 'logger' in list
+    assert 'logger_proxy' in list
   end
 
   test "Erlang root completion" do
     {:yes, '', list} = expand(':')
     assert is_list(list)
     assert 'lists' in list
+    assert 'Elixir.List' not in list
   end
 
   test "Elixir proxy" do
@@ -86,11 +80,6 @@ defmodule IEx.AutocompleteTest do
     assert expand('t(String.grap') == {:yes, 'heme', []}
   end
 
-  test "do not activate Elixir helper completion no operators" do
-    assert expand('t = String.co') == {:yes, '', ['codepoints/1', 'contains?/2']}
-    assert expand('t > String.grap') == {:yes, 'hemes', []}
-  end
-
   test "Elixir completion with self" do
     assert expand('Enumerable') == {:yes, '.', []}
   end
@@ -168,6 +157,7 @@ defmodule IEx.AutocompleteTest do
     assert expand('Xyz') == {:no, '', []}
     assert expand('x.Foo') == {:no, '', []}
     assert expand('x.Foo.get_by') == {:no, '', []}
+    assert expand('@foo.bar') == {:no, '', []}
   end
 
   test "Elixir root submodule completion" do
@@ -190,6 +180,25 @@ defmodule IEx.AutocompleteTest do
   test "function completion with arity" do
     assert expand('String.printable?') == {:yes, '', ['printable?/1', 'printable?/2']}
     assert expand('String.printable?/') == {:yes, '', ['printable?/1', 'printable?/2']}
+
+    assert expand('Enum.count') ==
+             {:yes, '', ['count/1', 'count/2', 'count_until/2', 'count_until/3']}
+
+    assert expand('Enum.count/') == {:yes, '', ['count/1', 'count/2']}
+  end
+
+  test "operator completion" do
+    assert expand('+') == {:yes, '', ['+/1', '+/2', '++/2']}
+    assert expand('+/') == {:yes, '', ['+/1', '+/2']}
+    assert expand('++/') == {:yes, '', ['++/2']}
+  end
+
+  test "sigil completion" do
+    {:yes, '', sigils} = expand('~')
+    assert '~C (sigil_C)' in sigils
+    {:yes, '', sigils} = expand('~r')
+    assert '"' in sigils
+    assert '(' in sigils
   end
 
   test "function completion using a variable bound to a module" do
@@ -228,21 +237,21 @@ defmodule IEx.AutocompleteTest do
     assert expand('map.f') == {:no, '', []}
   end
 
-  test "autocompletion off a bound variable only works for modules and maps" do
+  test "bound variables for modules and maps" do
     eval("num = 5; map = %{nested: %{num: 23}}")
     assert expand('num.print') == {:no, '', []}
     assert expand('map.nested.num.f') == {:no, '', []}
     assert expand('map.nested.num.key.f') == {:no, '', []}
   end
 
-  test "autocompletion using access syntax does is not supported" do
+  test "access syntax is not supported" do
     eval("map = %{nested: %{deeply: %{num: 23}}}")
     assert expand('map[:nested][:deeply].n') == {:no, '', []}
     assert expand('map[:nested].deeply.n') == {:no, '', []}
     assert expand('map.nested.[:deeply].n') == {:no, '', []}
   end
 
-  test "autocompletion off of unbound variables is not supported" do
+  test "unbound variables is not supported" do
     eval("num = 5")
     assert expand('other_var.f') == {:no, '', []}
     assert expand('a.b.c.d') == {:no, '', []}
@@ -276,11 +285,18 @@ defmodule IEx.AutocompleteTest do
   test "completion of manually imported functions and macros" do
     eval("import Enum; import Supervisor, only: [count_children: 1]; import Protocol")
 
-    functions_list = ['take/2', 'take_every/2', 'take_random/2', 'take_while/2']
-    assert expand('take') == {:yes, '', functions_list}
-
-    assert expand('count') == {:yes, '', ['count/1', 'count/2', 'count_children/1']}
     assert expand('der') == {:yes, 'ive', []}
+
+    assert expand('take') ==
+             {:yes, '', ['take/2', 'take_every/2', 'take_random/2', 'take_while/2']}
+
+    assert expand('take/') == {:yes, '', ['take/2']}
+
+    assert expand('count') ==
+             {:yes, '',
+              ['count/1', 'count/2', 'count_children/1', 'count_until/2', 'count_until/3']}
+
+    assert expand('count/') == {:yes, '', ['count/1', 'count/2']}
   end
 
   defmacro define_var do
@@ -304,23 +320,6 @@ defmodule IEx.AutocompleteTest do
     assert expand('{:zl') == {:yes, 'ib', []}
   end
 
-  test "ampersand completion" do
-    assert expand('&Enu') == {:yes, 'm', []}
-
-    functions_list = ['all?/1', 'all?/2', 'any?/1', 'any?/2', 'at/2', 'at/3']
-    assert expand('&Enum.a') == {:yes, [], functions_list}
-    assert expand('f = &Enum.a') == {:yes, [], functions_list}
-  end
-
-  test "negation operator completion" do
-    assert expand('!is_bin') == {:yes, 'ary', []}
-  end
-
-  test "pin operator completion" do
-    eval("my_variable = 2")
-    assert expand('^my_va') == {:yes, 'riable', []}
-  end
-
   defmodule SublevelTest.LevelA.LevelB do
   end
 
@@ -372,14 +371,53 @@ defmodule IEx.AutocompleteTest do
   end
 
   test "completion for struct names" do
-    assert expand('%IEx.AutocompleteTest.MyStr') == {:yes, 'uct', []}
+    assert {:yes, '', entries} = expand('%')
+    assert 'URI' in entries
+    assert 'IEx.History' in entries
+    assert 'IEx.State' in entries
+
+    assert {:yes, '', entries} = expand('%IEx.')
+    assert 'IEx.History' in entries
+    assert 'IEx.State' in entries
+
+    assert expand('%IEx.AutocompleteTe') == {:yes, 'st.MyStruct{', []}
+    assert expand('%IEx.AutocompleteTest.MyStr') == {:yes, 'uct{', []}
+
+    eval("alias IEx.AutocompleteTest.MyStruct")
+    assert expand('%MyStr') == {:yes, 'uct{', []}
   end
 
   test "completion for struct keys" do
+    assert {:yes, '', entries} = expand('%URI{')
+    assert 'path:' in entries
+    assert 'query:' in entries
+
+    assert {:yes, '', entries} = expand('%URI{path: "foo",')
+    assert 'path:' not in entries
+    assert 'query:' in entries
+
+    assert {:yes, 'ry: ', []} = expand('%URI{path: "foo", que')
+    assert {:no, [], []} = expand('%URI{path: "foo", unkno')
+    assert {:no, [], []} = expand('%Unkown{path: "foo", unkno')
+  end
+
+  test "completion for struct var keys" do
     eval("struct = %IEx.AutocompleteTest.MyStruct{}")
     assert expand('struct.my') == {:yes, '_val', []}
   end
 
+  test "completion for bitstring modifiers" do
+    assert {:yes, '', entries} = expand('<<foo::')
+    assert 'integer' in entries
+    assert 'size/1' in entries
+
+    assert {:yes, 'eger', []} = expand('<<foo::int')
+
+    assert {:yes, '', entries} = expand('<<foo::integer-')
+    refute 'integer' in entries
+    assert 'size/1' in entries
+  end
+
   test "ignore invalid Elixir module literals" do
     defmodule(:"Elixir.IEx.AutocompleteTest.Unicodé", do: nil)
     assert expand('IEx.AutocompleteTest.Unicod') == {:no, '', []}
@@ -387,4 +425,63 @@ defmodule IEx.AutocompleteTest do
     :code.purge(:"Elixir.IEx.AutocompleteTest.Unicodé")
     :code.delete(:"Elixir.IEx.AutocompleteTest.Unicodé")
   end
+
+  test "signature help for functions and macros" do
+    assert expand('String.graphemes(') == {:yes, '', ['graphemes(string)']}
+    assert expand('def ') == {:yes, '', ['def(call, expr \\\\ nil)']}
+
+    eval("import Enum; import Protocol")
+
+    assert ExUnit.CaptureIO.capture_io(fn ->
+             send(self(), expand('reduce('))
+           end) == "\nreduce(enumerable, acc, fun)"
+
+    assert_received {:yes, '', ['reduce(enumerable, fun)']}
+
+    assert expand('take(') == {:yes, '', ['take(enumerable, amount)']}
+    assert expand('derive(') == {:yes, '', ['derive(protocol, module, options \\\\ [])']}
+
+    defmodule NoDocs do
+      def sample(a), do: a
+    end
+
+    assert {:yes, [], [_ | _]} = expand('NoDocs.sample(')
+  end
+
+  @tag :tmp_dir
+  test "path completion inside strings", %{tmp_dir: dir} do
+    dir |> Path.join("single1") |> File.touch()
+    dir |> Path.join("file1") |> File.touch()
+    dir |> Path.join("file2") |> File.touch()
+    dir |> Path.join("dir") |> File.mkdir()
+    dir |> Path.join("dir/file3") |> File.touch()
+    dir |> Path.join("dir/file4") |> File.touch()
+
+    assert expand('"./') == path_autocompletion(".")
+    assert expand('"/') == path_autocompletion("/")
+    assert expand('"./#\{') == expand('{')
+    assert expand('"./#\{Str') == expand('{Str')
+    assert expand('Path.join("./", is_') == expand('is_')
+
+    assert expand('"#{dir}/') == path_autocompletion(dir)
+    assert expand('"#{dir}/sin') == {:yes, 'gle1', []}
+    assert expand('"#{dir}/single1') == {:yes, '"', []}
+    assert expand('"#{dir}/fi') == {:yes, 'le', []}
+    assert expand('"#{dir}/file') == path_autocompletion(dir, "file")
+    assert expand('"#{dir}/d') == {:yes, 'ir/', []}
+    assert expand('"#{dir}/dir') == {:yes, '/', []}
+    assert expand('"#{dir}/dir/') == {:yes, 'file', []}
+    assert expand('"#{dir}/dir/file') == dir |> Path.join("dir") |> path_autocompletion("file")
+  end
+
+  defp path_autocompletion(dir, hint \\ "") do
+    dir
+    |> File.ls!()
+    |> Stream.filter(&String.starts_with?(&1, hint))
+    |> Enum.map(&String.to_charlist/1)
+    |> case do
+      [] -> {:no, '', []}
+      list -> {:yes, '', list}
+    end
+  end
 end
