@@ -18,7 +18,7 @@ defmodule Mix.Tasks.Compile do
       which are `[:yecc, :leex, :erlang, :elixir, :app]`.
 
     * `:consolidate_protocols` - when `true`, runs protocol
-      consolidation via the `compile.protocols` task. The default
+      consolidation via the `mix compile.protocols` task. The default
       value is `true`.
 
     * `:build_embedded` - when `true`, embeds all code and priv
@@ -46,17 +46,22 @@ defmodule Mix.Tasks.Compile do
 
   ## Command line options
 
+    * `--all-warnings` - prints warnings even from files that do not need to be recompiled
     * `--erl-config` - path to an Erlang term file that will be loaded as Mix config
     * `--force` - forces compilation
     * `--list` - lists all enabled compilers
-    * `--no-app-loading` - does not load applications (including from deps) before compiling
+    * `--no-app-loading` - does not load .app resource file after compilation
     * `--no-archives-check` - skips checking of archives
     * `--no-compile` - does not actually compile, only loads code and perform checks
     * `--no-deps-check` - skips checking of dependencies
     * `--no-elixir-version-check` - does not check Elixir version
+    * `--no-optional-deps` - does not compile or load optional deps. Useful for testing
+      if a library still successfully compiles without optional dependencies (which is the
+      default case with dependencies)
     * `--no-protocol-consolidation` - skips protocol consolidation
     * `--no-validate-compile-env` - does not validate the application compile environment
     * `--return-errors` - returns error status and diagnostics instead of exiting on error
+    * `--warnings-as-errors` - exit with non-zero status if compilation has one or more warnings
 
   """
 
@@ -64,9 +69,17 @@ defmodule Mix.Tasks.Compile do
   Returns all compilers.
   """
   def compilers(config \\ Mix.Project.config()) do
-    # TODO: Deprecate :xref on v1.12
     compilers = config[:compilers] || Mix.compilers()
-    List.delete(compilers, :xref)
+
+    if :xref in compilers do
+      IO.warn(
+        "the :xref compiler is deprecated, please remove it from your mix.exs :compilers options"
+      )
+
+      List.delete(compilers, :xref)
+    else
+      compilers
+    end
   end
 
   @impl true
@@ -102,7 +115,8 @@ defmodule Mix.Tasks.Compile do
       shell.info(format('mix ~-#{max}s # ~ts', [task, doc]))
     end)
 
-    compilers = compilers() ++ if(consolidate_protocols?(:ok), do: [:protocols], else: [])
+    consolidate_protocols? = Mix.Project.config()[:consolidate_protocols]
+    compilers = compilers() ++ if(consolidate_protocols?, do: [:protocols], else: [])
     shell.info("\nEnabled compilers: #{Enum.join(compilers, ", ")}")
     :ok
   end
@@ -115,6 +129,8 @@ defmodule Mix.Tasks.Compile do
     {opts, _, _} = OptionParser.parse(args, switches: [erl_config: :string])
     load_erl_config(opts)
 
+    # We invoke compile.all even on --no-compile because compile.all
+    # loads the apps, and we still want to do that.
     {res, diagnostics} =
       Mix.Task.run("compile.all", args)
       |> List.wrap()
@@ -123,17 +139,31 @@ defmodule Mix.Tasks.Compile do
 
     config = Mix.Project.config()
 
-    cond do
-      "--no-compile" in args ->
-        Mix.Task.reenable("compile")
-        {:noop, []}
+    consolidate_protocols? =
+      config[:consolidate_protocols] and "--no-protocol-consolidation" not in args
 
-      config[:consolidate_protocols] and "--no-protocol-consolidation" not in args ->
-        {consolidate_and_load_protocols(args, config, res), diagnostics}
+    res =
+      cond do
+        "--no-compile" in args ->
+          Mix.Task.reenable("compile")
+          :noop
 
-      true ->
-        {res, diagnostics}
+        consolidate_protocols? and reconsolidate_protocols?(res) ->
+          Mix.Task.run("compile.protocols", args)
+          :ok
+
+        true ->
+          res
+      end
+
+    with true <- consolidate_protocols?,
+         path = Mix.Project.consolidation_path(config),
+         {:ok, protocols} <- File.ls(path) do
+      Code.prepend_path(path)
+      Enum.each(protocols, &load_protocol/1)
     end
+
+    {res, diagnostics}
   end
 
   defp format(expression, args) do
@@ -177,28 +207,9 @@ defmodule Mix.Tasks.Compile do
 
   ## Consolidation handling
 
-  defp consolidate_protocols?(:ok), do: true
-  defp consolidate_protocols?(:noop), do: not Mix.Tasks.Compile.Protocols.consolidated?()
-  defp consolidate_protocols?(:error), do: false
-
-  defp consolidate_and_load_protocols(args, config, res) do
-    res =
-      if consolidate_protocols?(res) do
-        Mix.Task.run("compile.protocols", args)
-        :ok
-      else
-        res
-      end
-
-    path = Mix.Project.consolidation_path(config)
-
-    with {:ok, protocols} <- File.ls(path) do
-      Code.prepend_path(path)
-      Enum.each(protocols, &load_protocol/1)
-    end
-
-    res
-  end
+  defp reconsolidate_protocols?(:ok), do: true
+  defp reconsolidate_protocols?(:noop), do: not Mix.Tasks.Compile.Protocols.consolidated?()
+  defp reconsolidate_protocols?(:error), do: false
 
   defp load_protocol(file) do
     case file do
