@@ -1,6 +1,9 @@
 defmodule Mix.Utils do
   @moduledoc false
 
+  # For bootstrapping purposes
+  @compile {:no_warn_undefined, Logger}
+
   @doc """
   Gets the Mix home.
 
@@ -32,6 +35,24 @@ defmodule Mix.Utils do
     mix_home_xdg_lookup(:user_config)
   end
 
+  @doc """
+  Gets possible location of Mix cache.
+
+  Possible locations:
+
+   * `XDG_CACHE_HOME/mix` (if `MIX_XDG` is set)
+   * `:filename.basedir(:user_cache, "mix")`
+
+  """
+  def mix_cache do
+    if System.get_env("MIX_XDG") in ["1", "true"] do
+      # XDG lookups are only done for linux OS
+      :filename.basedir(:user_cache, "mix", %{os: :linux})
+    else
+      :filename.basedir(:user_cache, "mix")
+    end
+  end
+
   defp mix_home_xdg_lookup(xdg) do
     cond do
       dir = System.get_env("MIX_HOME") ->
@@ -414,10 +435,28 @@ defmodule Mix.Utils do
     |> Enum.map_join(".", &Macro.camelize/1)
   end
 
+  @doc """
+  Symlinks or copy with the option to force a hard copy.
+  See `symlink_or_copy/2`.
+  """
+  def symlink_or_copy(hard_copy?, source, target) do
+    if hard_copy? do
+      if File.exists?(source) do
+        File.rm_rf!(target)
+        File.cp_r!(source, target)
+      end
+    else
+      symlink_or_copy(source, target)
+    end
+  end
+
   @doc """
   Symlinks directory `source` to `target` or copies it recursively
   in case symlink fails.
 
+  In case of conflicts, it copies files only if they have been
+  recently touched.
+
   Expects source and target to be absolute paths as it generates
   a relative symlink.
   """
@@ -429,14 +468,32 @@ defmodule Mix.Utils do
           {:win32, _} -> source
           _ -> make_relative_path(source, target)
         end
-        |> String.to_charlist()
 
-      case :file.read_link(target) do
+      case File.read_link(target) do
         {:ok, ^link} ->
           :ok
 
         {:ok, _} ->
-          File.rm!(target)
+          case File.rm(target) do
+            :ok ->
+              :ok
+
+            {:error, reason} ->
+              reason = IO.iodata_to_binary(:file.format_error(reason))
+
+              Mix.raise("""
+              Cannot remove symlink #{inspect(target)} due to reason: #{reason}"
+
+                * Make sure you have permission to access the _build directory
+                  (you may have the wrong permission if you change users or ran as admin)
+
+                * If you are using Windows, avoid using substitute drives,
+                  as they don't play well with symlinks
+
+                * In case the issue continues, consider removing the _build directory
+              """)
+          end
+
           do_symlink_or_copy(source, target, link)
 
         {:error, :enoent} ->
@@ -455,17 +512,21 @@ defmodule Mix.Utils do
   end
 
   defp do_symlink_or_copy(source, target, link) do
-    case :file.make_symlink(link, target) do
+    case File.ln_s(link, target) do
       :ok ->
         :ok
 
       {:error, _} ->
-        file =
-          File.cp_r!(source, target, fn orig, dest ->
-            File.stat!(orig).mtime > File.stat!(dest).mtime
-          end)
+        files =
+          File.cp_r!(source, target,
+            on_conflict: fn orig, dest ->
+              {orig_mtime, orig_size} = last_modified_and_size(orig)
+              {dest_mtime, dest_size} = last_modified_and_size(dest)
+              orig_mtime > dest_mtime or orig_size != dest_size
+            end
+          )
 
-        {:ok, file}
+        {:ok, files}
     end
   end
 
@@ -593,25 +654,51 @@ defmodule Mix.Utils do
     headers = [{'user-agent', 'Mix/#{System.version()}'}]
     request = {:binary.bin_to_list(path), headers}
 
+    # Use the system certificates if available, otherwise skip peer verification
+    # TODO: Always use system certificates when OTP >= 25.1 is required
+    ssl_options =
+      if Code.ensure_loaded?(:httpc) and function_exported?(:httpc, :ssl_verify_host_options, 1) do
+        try do
+          apply(:httpc, :ssl_verify_host_options, [true])
+        rescue
+          _ ->
+            Mix.shell().error(
+              "warning: failed to load system certificates. SSL peer verification will be skipped but downloads are still verified with a checksum"
+            )
+
+            [verify: :verify_none]
+        end
+      else
+        [verify: :verify_none]
+      end
+
     # We are using relaxed: true because some servers is returning a Location
     # header with relative paths, which does not follow the spec. This would
     # cause the request to fail with {:error, :no_scheme} unless :relaxed
     # is given.
     #
     # If a proxy environment variable was supplied add a proxy to httpc.
-    http_options = [relaxed: true] ++ proxy_config(path)
+    http_options = [relaxed: true, ssl: ssl_options] ++ proxy_config(path)
 
-    case httpc_request(request, http_options) do
-      {:error, {:failed_connect, [{:to_address, _}, {inet, _, reason}]}}
-      when inet in [:inet, :inet6] and reason in [:ehostunreach, :enetunreach] ->
-        :httpc.set_options([ipfamily: fallback(inet)], :mix)
-        request |> httpc_request(http_options) |> httpc_response()
+    # Silence the warning from OTP as we verify the contents
+    level = Logger.level()
+    Logger.configure(level: :error)
 
-      response ->
-        httpc_response(response)
+    try do
+      case httpc_request(request, http_options) do
+        {:error, {:failed_connect, [{:to_address, _}, {inet, _, reason}]}}
+        when inet in [:inet, :inet6] and
+               reason in [:ehostunreach, :enetunreach, :eprotonosupport, :nxdomain] ->
+          :httpc.set_options([ipfamily: fallback(inet)], :mix)
+          request |> httpc_request(http_options) |> httpc_response()
+
+        response ->
+          httpc_response(response)
+      end
+    after
+      Logger.configure(level: level)
+      :inets.stop(:httpc, :mix)
     end
-  after
-    :inets.stop(:httpc, :mix)
   end
 
   defp fallback(:inet), do: :inet6
