@@ -23,7 +23,7 @@ defmodule Mix.Tasks.TestTest do
       assert ex_unit_opts_from_given(only: "focus", include: "special") == only
     end
 
-    test "translates :color into list containing an enabled key/value pair" do
+    test "translates :color into list containing an enabled key-value pair" do
       assert ex_unit_opts_from_given(color: false) == [colors: [enabled: false]]
       assert ex_unit_opts_from_given(color: true) == [colors: [enabled: true]]
     end
@@ -32,9 +32,14 @@ defmodule Mix.Tasks.TestTest do
       assert ex_unit_opts_from_given(formatter: "A.B") == [formatters: [A.B]]
     end
 
+    test "accepts custom :exit_status" do
+      assert {:exit_status, 5} in ex_unit_opts(exit_status: 5)
+    end
+
     test "includes some default options" do
       assert ex_unit_opts([]) == [
                autorun: false,
+               exit_status: 2,
                failures_manifest_file:
                  Path.join(Mix.Project.manifest_path(), ".mix_test_failures")
              ]
@@ -48,7 +53,7 @@ defmodule Mix.Tasks.TestTest do
     defp ex_unit_opts_from_given(passed) do
       passed
       |> ex_unit_opts()
-      |> Keyword.drop([:failures_manifest_file, :autorun])
+      |> Keyword.drop([:failures_manifest_file, :autorun, :exit_status])
     end
   end
 
@@ -68,12 +73,12 @@ defmodule Mix.Tasks.TestTest do
         assert_stale_run_output("2 tests, 0 failures")
 
         set_all_mtimes()
-        File.touch!("lib/b.ex")
+        force_recompilation("lib/b.ex")
 
         assert_stale_run_output("1 test, 0 failures")
 
         set_all_mtimes()
-        File.touch!("lib/a.ex")
+        force_recompilation("lib/a.ex")
 
         assert_stale_run_output("2 tests, 0 failures")
       end)
@@ -131,7 +136,10 @@ defmodule Mix.Tasks.TestTest do
   describe "--cover" do
     test "reports the coverage of each app's modules in an umbrella" do
       in_fixture("umbrella_test", fn ->
-        output = mix(["test", "--cover"])
+        # This fixture by default results in coverage above the default threshold
+        # which should result in an exit status of 0.
+        assert {output, 0} = mix_code(["test", "--cover"])
+        assert output =~ "4 tests, 0 failures"
 
         # For bar, we do regular --cover and also test protocols
         assert output =~ """
@@ -145,6 +153,8 @@ defmodule Mix.Tasks.TestTest do
                   100.00% | Total
                """
 
+        assert output =~ "1 test, 0 failures"
+
         # For foo, we do regular --cover and test it does not include bar
         assert output =~ """
                Generating cover results ...
@@ -155,6 +165,19 @@ defmodule Mix.Tasks.TestTest do
                -----------|--------------------------
                   100.00% | Total
                """
+
+        # We skip a test in bar to force coverage below the default threshold
+        # which should result in an exit status of 1.
+        assert {output, code} = mix_code(["test", "--cover", "--exclude", "maybe_skip"])
+
+        assert output =~ """
+               Coverage test failed, threshold not met:
+
+                   Coverage:    0.00%
+                   Threshold:  90.00%
+               """
+
+        assert code == 3
       end)
     end
 
@@ -170,6 +193,7 @@ defmodule Mix.Tasks.TestTest do
                Percentage | Module
                -----------|--------------------------
                   100.00% | Bar
+                  100.00% | Bar.Ignore
                   100.00% | Bar.Protocol
                   100.00% | Bar.Protocol.BitString
                   100.00% | Foo
@@ -235,6 +259,7 @@ defmodule Mix.Tasks.TestTest do
       end)
     end
 
+    @tag :unix
     test "does not exit on compilation failure" do
       in_fixture("test_stale", fn ->
         File.write!("lib/b.ex", """
@@ -334,13 +359,62 @@ defmodule Mix.Tasks.TestTest do
         )
       end)
     end
+
+    test "do not raise if partitions flag is set to 1 and no partition given" do
+      in_fixture("test_stale", fn ->
+        assert mix(["test", "--partitions", "1"], []) =~
+                 "2 tests, 0 failures"
+
+        assert mix(["test", "--partitions", "1"], [{"MIX_TEST_PARTITION", ""}]) =~
+                 "2 tests, 0 failures"
+
+        assert mix(["test", "--partitions", "1"], [{"MIX_TEST_PARTITION", "1"}]) =~
+                 "2 tests, 0 failures"
+      end)
+    end
+
+    test "raise if partitions is set to non-positive value" do
+      in_fixture("test_stale", fn ->
+        File.write!("test/test_helper.exs", """
+        Mix.shell(Mix.Shell.Process)
+        ExUnit.start()
+        """)
+
+        assert_run_output(
+          ["--partitions", "0"],
+          "--partitions : expected to be positive integer, got 0"
+        )
+
+        assert_run_output(
+          ["--partitions", "-1"],
+          "--partitions : expected to be positive integer, got -1"
+        )
+      end)
+    end
+
+    test "runs after_suite with partitions with no tests" do
+      in_fixture("test_stale", fn ->
+        File.write!("test/test_helper.exs", """
+        ExUnit.after_suite(fn _stats -> IO.puts("AFTER SUITE") end)
+        ExUnit.start()
+        """)
+
+        assert mix(["test", "--partitions", "3"], [{"MIX_TEST_PARTITION", "3"}]) =~ """
+               AFTER SUITE
+               There are no tests to run
+               """
+      end)
+    end
   end
 
   describe "logs and errors" do
     test "logs test absence for a project with no test paths" do
       in_fixture("test_stale", fn ->
         File.rm_rf!("test")
+        assert_run_output("There are no tests to run")
 
+        File.mkdir_p!("test")
+        File.write!("test/test_helper.exs", "ExUnit.start()")
         assert_run_output("There are no tests to run")
       end)
     end
@@ -375,9 +449,14 @@ defmodule Mix.Tasks.TestTest do
       in_fixture("umbrella_test", fn ->
         # Run false positive test first so at least the code is compiled
         # and we can perform more aggressive assertions later
-        assert mix(["test", "apps/unknown_app/test"]) =~ """
+        output = mix(["test", "apps/unknown_app/test"])
+
+        assert output =~ """
                ==> bar
                Paths given to "mix test" did not match any directory/file: apps/unknown_app/test
+               """
+
+        assert output =~ """
                ==> foo
                Paths given to "mix test" did not match any directory/file: apps/unknown_app/test
                """
@@ -386,7 +465,17 @@ defmodule Mix.Tasks.TestTest do
 
         assert output =~ """
                ==> bar
-               .
+               ....
+               """
+
+        refute output =~ "==> foo"
+        refute output =~ "Paths given to \"mix test\" did not match any directory/file"
+
+        output = mix(["test", "./apps/bar/test/bar_tests.exs"])
+
+        assert output =~ """
+               ==> bar
+               ....
                """
 
         refute output =~ "==> foo"
@@ -408,6 +497,66 @@ defmodule Mix.Tasks.TestTest do
     end
   end
 
+  describe "--warnings-as-errors" do
+    test "fail on warning in tests" do
+      in_fixture("test_stale", fn ->
+        msg =
+          "Test suite aborted after successful execution due to warnings while using the --warnings-as-errors option"
+
+        refute mix(["test", "--warnings-as-errors"]) =~ msg
+
+        File.write!("lib/warning.ex", """
+        unused_compile_var = 1
+        """)
+
+        File.write!("test/warning_test_stale.exs", """
+        defmodule WarningTest do
+          use ExUnit.Case
+
+          test "warning" do
+            unused_test_var = 1
+          end
+        end
+        """)
+
+        output = mix(["test", "--warnings-as-errors", "test/warning_test_stale.exs"])
+        assert output =~ "variable \"unused_compile_var\" is unused"
+        assert output =~ "variable \"unused_test_var\" is unused"
+        assert output =~ msg
+      end)
+    end
+
+    test "mark failed tests" do
+      in_fixture("test_failed", fn ->
+        File.write!("test/warning_test_failed.exs", """
+        defmodule WarningTest do
+          use ExUnit.Case
+
+          test "warning" do
+            unused_var = 123
+          end
+        end
+        """)
+
+        output = mix(["test", "--warnings-as-errors"])
+        assert output =~ "2 failures"
+        refute output =~ "Test suite aborted after successful execution"
+        output = mix(["test", "--failed"])
+        assert output =~ "2 failures"
+      end)
+    end
+  end
+
+  describe "--exit-status" do
+    test "returns custom exit status" do
+      in_fixture("test_failed", fn ->
+        {output, exit_status} = mix_code(["test", "--exit-status", "5"])
+        assert output =~ "2 failures"
+        assert exit_status == 5
+      end)
+    end
+  end
+
   defp receive_until_match(port, expected, acc) do
     receive do
       {^port, {:data, output}} ->
